"""Top-level package for aiolifxc."""

from .aiolifx import Devices, Lights, Device, Light, LifxDiscovery, DeviceOffline  # NOQA
from .colors import Color  # NOQA

__author__ = """Brian May"""
__email__ = 'brian@linuxpenguins.xyz'
__version__ = '0.5.3'



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# This application is simply a bridge application for Lifx bulbs.
#
# Copyright (c) 2016 François Wautier
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies
# or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
# IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
import asyncio as aio
from collections import Awaitable
import datetime
import logging
import random
import socket
from typing import (
    List, Dict, Type, Optional, Union, Tuple, Callable, TypeVar,
    Any, AnyStr, cast)
from typing import Set  # NOQA

from .colors import Color
from .message import Message, BROADCAST_MAC
from . import msgtypes
from .products import product_map
from .unpack import unpack_lifx_message

# A couple of constants
UDP_BROADCAST_IP = "255.255.255.255"
UDP_BROADCAST_PORT = 56700
DEFAULT_TIMEOUT = 0.5  # How long to wait for an ack or response
DEFAULT_ATTEMPTS = 3  # How many time should we try to send to the bulb`
DISCOVERY_INTERVAL = 180
DISCOVERY_STEP = 5

GenericResponse = TypeVar('GenericResponse', bound=Message)
GenericDevices = TypeVar('GenericDevices', bound='Devices')
GenericDevice = TypeVar('GenericDevice', bound='Device')
Power = Union[bool, int]

logger = logging.getLogger(__name__)


def _mac_to_ipv6_link_local(mac: str, prefix: str) -> str:
    """ Translate a MAC address into an IPv6 address in the prefixed network"""

    # Remove the most common delimiters; dots, dashes, etc.
    mac_value = int(mac.translate(str.maketrans(dict([(x, None) for x in [" ", ".", ":", "-"]]))), 16)
    # Split out the bytes that slot into the IPv6 address
    # XOR the most significant byte with 0x02, inverting the
    # Universal / Local bit
    high2 = mac_value >> 32 & 0xffff ^ 0x0200
    high1 = mac_value >> 24 & 0xff
    low1 = mac_value >> 16 & 0xff
    low2 = mac_value & 0xffff
    return prefix + ':{:04x}:{:02x}ff:fe{:02x}:{:04x}'.format(
        high2, high1, low1, low2)


def _nanosec_to_hours(ns: int) -> float:
    return ns / (1000000000.0 * 60 * 60)


def _str_map(key: Optional[Power]) -> str:
    string_representation = "Unknown (%s)" % key
    if key is None:
        string_representation = "Unknown"
    elif isinstance(key, bool):
        if key is True:
            string_representation = "On"
        elif key is False:
            string_representation = "Off"
    return string_representation


class DeviceOffline(Exception):
    pass


class Devices:
    """ Class that represents a number of Devices. """
    def __init__(self, loop: aio.AbstractEventLoop) -> None:
        self._device_list = []  # type: List[Device]
        self._loop = loop
        self._all = True
        self._parent = None  # type: Optional['Devices']
        self._groups = None  # type: Optional[List[str]]
        self._labels = None  # type: Optional[List[str]]
        self._mac_addrs = None  # type: Optional[List[str]]

    def register(self, device: 'Device') -> None:
        """
        Add a new device.

        :param device: The device to add.
        """
        self._loop.create_task(self.async_register(device))

    async def async_register(self, device: 'Device') -> None:
        try:
            await device.get_metadata(loop=self._loop)
            if self._parent is None:
                self._device_list.append(device)
            logger.info("Registered light %s.", device)
        except DeviceOffline:
            logger.error("Light is offline %s", device)

    def unregister(self, device: 'Device') -> None:
        """
        Remove an existing device.

        :param device: The light to remove.
        """
        logger.info("Unregistered light %s.", device)
        if self._parent is None:
            idx = 0
            for x in list([y.mac_addr for y in self._device_list]):
                if x == device.mac_addr:
                    del(self._device_list[idx])
                    break
                idx += 1

    def get_list(self, device_class: Type[GenericDevice]) -> List[GenericDevice]:
        if self._parent is not None:
            device_list = self._parent.get_list(Device)
        else:
            device_list = self._device_list

        devices = set()  # type: Set['Device']

        # Filter list according to requirements of this object
        if self._all:
            devices |= set([
                d for d in device_list
            ])

        if self._groups is not None:
            for group in self._groups:
                devices |= set([
                    d for d in device_list if d.group == group
                ])

        if self._labels is not None:
            for label in self._labels:
                devices |= set([
                    d for d in device_list if d.label == label
                ])

        if self._mac_addrs is not None:
            for mac_addr in self._mac_addrs:
                devices |= set([
                    d for d in device_list if d.mac_addr == mac_addr
                ])

        # Ensure all objects are subclasses of GenericDevice.
        result = []  # type: List[GenericDevice]
        for device in devices:
            if isinstance(device, device_class):
                result.append(device)

        # Return the result
        return result

    @property
    def device_list(self) -> List['Device']:
        return self.get_list(Device)

    def start_discover(
            self, ipv6prefix: str=None,
            discovery_interval: int=DISCOVERY_INTERVAL,
            discovery_step: int=DISCOVERY_STEP) -> aio.Task:
        """
        Get the Task that will discoveries.

        :param ipv6prefix: The IPv6 prefix to use for IPv6 addresses.
        :param discovery_interval: How often should we rerun discover (seconds)?
        :param discovery_step: How often should we wake up (seconds)?
        :return: The asyncio task.
        """
        def lifx_discovery() -> aio.BaseProtocol:
            """ Construct an LIFX discovery protocol handler. """
            return LifxDiscovery(
                loop=self._loop,
                devices=self,
                ipv6prefix=ipv6prefix,
                discovery_interval=discovery_interval,
                discovery_step=discovery_step,
            )

        coro = self._loop.create_datagram_endpoint(
            lifx_discovery,
            local_addr=('0.0.0.0', UDP_BROADCAST_PORT),
        )
        return self._loop.create_task(coro)

    def get_clone(self, new_class: Type[GenericDevices]) -> GenericDevices:
        """
        Get clone Devices object.

        :param new_class: Class to use to create new object.
        :return: The new object.

        The clone will reference the objects from the parent.
        """
        # noinspection PyCallingNonCallable
        child = new_class(loop=self._loop)
        child._parent = self
        return child

    def get_by_group(self: GenericDevices, group: str) -> GenericDevices:
        """
        Get clone Devices object filtered by group.

        :param group: The name of the group.
        :return: The new object.

        The groups must be loaded already in the devices.
        """
        result = self.get_clone(type(self))
        result._groups = [group]
        result._all = False
        return result

    def get_by_label(self: GenericDevices, label: str) -> GenericDevices:
        """
        Get a clone Devices object filtered by label.

        :param label: The name of the label.
        :return: The new object.

        The labels must be loaded already in the devices.
        """
        result = self.get_clone(type(self))
        result._labels = [label]
        result._all = False
        return result

    def get_by_mac_addr(self: GenericDevices, mac_addr: str) -> GenericDevices:
        """
        Get a clone Devices object filtered by label.

        :param label: The name of the label.
        :return: The new object.

        The labels must be loaded already in the devices.
        """
        result = self.get_clone(type(self))
        result._mac_addrs = [mac_addr]
        result._all = False
        return result

    def get_by_lists(
            self: GenericDevices, *,
            groups: List[str]=None,
            labels: List[str]=None,
            mac_addrs: List[str]=None) -> GenericDevices:
        """
        Get a clone Devices object filtered by group list and label list.
        :param groups: The list of groups for the filter.
        :param labels: The list of labels for the filter.
        :return: The new object.

        The groups and labels must be loaded already in the devices.

        """
        result = self.get_clone(type(self))
        result._groups = groups
        result._labels = labels
        result._mac_addrs = mac_addrs
        result._all = False
        return result

    async def do_for_every_device(
            self: GenericDevices,
            device_class: Type[GenericDevice],
            fun: Callable[[GenericDevice], Awaitable],
    ) -> None:
        """
        Run a async function for every device.

        :param fun: The function to call.

        Errors will get logged but not propagated.
        """
        async def wrapper(device: GenericDevice) -> None:
            try:
                await fun(device)
            except DeviceOffline:
                logger.info("Light is offline %s", device)
            except Exception:
                logger.exception(
                    "An exception was generated in do_for_every_device for %s:",
                    device,
                )

        coroutines = []
        for device in self.get_list(device_class):
            coroutines.append(wrapper(device))
        await aio.gather(*coroutines, loop=self._loop)

    async def get_meta_information(self) -> None:
        """ Get all meta information for devices. """
        async def single_device(device: Device) -> None:
            await device.get_metadata(loop=self._loop)
        await self.do_for_every_device(Device, single_device)

    async def set_power(self, value: Power, rapid: bool=False) -> None:
        """ Set power for all devices. """
        async def single_device(device: Device) -> None:
            await device.set_power(value=value, rapid=rapid)
        await self.do_for_every_device(Device, single_device)

    def __str__(self) -> str:
        return format(", ".join([str(d) for d in self.device_list]))


class Lights(Devices):
    """ Class that represents a number of Lights. """
    def __init__(self, loop: aio.AbstractEventLoop) -> None:
        super().__init__(loop=loop)
        self._device_list = []  # type: List[Light]

    @property
    def light_list(self) -> List['Light']:
        return self.get_list(Light)

    async def set_light_power(self, value: Power, duration: int=0, rapid: bool=False) -> None:
        """ Set power for all lights. """
        async def single_device(device: Light) -> None:
            await device.set_light_power(value=value, duration=duration, rapid=rapid)
        await self.do_for_every_device(Light, single_device)

    async def set_color(self, color: Color, duration: int = 0, rapid: bool = False) -> None:
        """ Set color for all lights. """
        async def single_device(device: Light) -> None:
            await device.set_color(color=color, duration=duration, rapid=rapid)
        await self.do_for_every_device(Light, single_device)

    async def set_waveform(
            self, *,
            color: Color,
            transient: int, period: int, cycles: int, duty_cycle: int, waveform: int,
            rapid: bool = False) -> None:
        """ Set waveform for all lights. """
        async def single_device(device: Light) -> None:
            await device.set_waveform(
                color=color,
                transient=transient, period=period, cycles=cycles, duty_cycle=duty_cycle, waveform=waveform,
                rapid=rapid)

        await self.do_for_every_device(Light, single_device)


class Device(aio.DatagramProtocol):
    """ Implement common functions for a LIFX Device. """

    # mac_addr is a string, with the ":" and everything.
    # ip_addr is a string with the ip address
    # port is the port we are connected to
    def __init__(
            self, *, loop: aio.AbstractEventLoop,
            mac_addr: str, ip_addr: str, port: int, devices: Devices=None) -> None:
        """
        Construct a new Device object.

        :param loop: The Asyncio event loop.
        :param mac_addr: The MAC Address. with the ":" and everything.
        :param ip_addr: A string with the IP address.
        :param port: The UDP port to use.
        :param devices: The lights list this light belongs to.
        """
        self._loop = loop
        self._mac_addr = mac_addr
        self._ip_addr = ip_addr
        self._port = port
        self._devices = devices
        self._registered = False  # type: bool
        self._retry_count = DEFAULT_ATTEMPTS
        self._timeout = DEFAULT_TIMEOUT
        self._transport = None  # type: Optional[aio.DatagramTransport]
        self._task = None  # type: Optional[aio.Task]
        self._seq = 0
        # Key is the message sequence, value is (response type, Event, response)
        self._message = {}  # type: Dict[int, List]
        self._source_id = random.randint(0, (2 ** 32) - 1)
        # Default callback for unexpected messages
        self._default_callback = None  # type: Optional[Callable[[Message], None]]
        # And the rest
        self._label = None  # type: Optional[str]
        self._location = None  # type: Optional[str]
        self._group = None  # type: Optional[str]
        self._power_level = None  # type: Optional[Power]
        self._vendor = None  # type: Optional[int]
        self._product = None  # type: Optional[int]
        self._version = None  # type: Optional[int]
        self._host_firmware_version = None  # type: Optional[str]
        self._host_firmware_build_timestamp = None  # type: Optional[int]
        self._wifi_firmware_version = None  # type: Optional[str]
        self._wifi_firmware_build_timestamp = None  # type: Optional[int]
        self._last_msg = datetime.datetime.now() - datetime.timedelta(seconds=600)

    @property
    def mac_addr(self) -> str:
        """ Return the MAC address associated with this device. """
        return self._mac_addr

    @property
    def label(self) -> str:
        """ Return the cached label - if any - for this device. """
        if self._label is None:
            raise RuntimeError("Label has not been loaded")
        return self._label

    @property
    def group(self) -> str:
        """ Return the cached group - if any - for this device. """
        if self._group is None:
            raise RuntimeError("Group has not been loaded")
        return self._group

    @property
    def ip_addr(self) -> str:
        """ Return the MAC address associated with this device. """
        return self._ip_addr

    def _seq_next(self) -> int:
        self._seq = (self._seq + 1) % 128
        return self._seq

    #
    #                            Protocol Methods
    #

    def connection_made(self, transport: aio.BaseTransport) -> None:
        """ Called when a connection is made. """
        self._transport = cast(aio.DatagramTransport, transport)
        self.register()

    def datagram_received(self, data: AnyStr, addr: str) -> None:
        """ Called when we receive a packet. """
        self.register()
        assert isinstance(data, bytes)
        response = unpack_lifx_message(data)
        if response.seq_num in self._message:
            self._last_msg = datetime.datetime.now()
            response_type, myevent, __ = self._message[response.seq_num]
            if type(response) == response_type:
                if response.source_id == self._source_id:
                    self._message[response.seq_num][2] = response
                    myevent.set()
        elif self._default_callback:
            self._default_callback(response)

    def register(self) -> None:
        """ Register this device to `Devices` object. """
        if not self._registered:
            self._registered = True
            if self._devices:
                self._devices.register(self)

    def unregister(self) -> None:
        """ Unregister this device to `Devices` object. """
        if self._registered:
            # Only if we have not received any message recently.
            # On slower CPU, a race condition seem to sometime occur
            if datetime.datetime.now() - datetime.timedelta(seconds=DEFAULT_TIMEOUT) > self._last_msg:
                self._registered = False
                if self._devices:
                    self._devices.unregister(self)

    def renew(self, *, family: int, ip_addr: str, port: int) -> None:
        """
        Renew the light registration with updated contact information.

        :param ip_addr: A string with the IP address.
        :param port: The UDP port to use.
        """
        if self._ip_addr != ip_addr or self._port != port:
            self.cleanup()
            self._ip_addr = ip_addr
            self._port = port

        if self._task is None:
            coro = self._loop.create_datagram_endpoint(
                lambda: self, family=family, remote_addr=(self._ip_addr, self._port))
            self._task = self._loop.create_task(coro)

    def cleanup(self) -> None:
        """ Cleanup all resources used by this `Device` object. """
        if self._transport:
            self._transport.close()
            self._transport = None
        if self._task:
            self._task.cancel()
            self._task = None

    #
    #                            Workflow Methods
    #

    async def _fire_sending(self, msg: Message, num_repeats: int) -> None:
        """
        Send a message a number of times.

        :param msg: The message to send.
        :param num_repeats: The number of times we should send it.
        """
        assert self._transport is not None

        if num_repeats is None:
            num_repeats = self._retry_count
        sent_msg_count = 0
        sleep_interval = 0.05
        while sent_msg_count < num_repeats:
            packed_message = msg.generate_packed_message()
            self._transport.sendto(packed_message)
            sent_msg_count += 1
            # Max num of messages device can handle is 20 per second.
            await aio.sleep(sleep_interval)

    def _fire_and_forget(
            self, msg_type: Type[Message], payload: Dict[str, Any]=None,
            *, num_repeats: int=1) -> None:
        """
        Don't wait for Acks or Responses, just send the same message repeatedly as fast as
        possible in a separate task.

        :param msg_type: The type of the Message.
        :param payload: The payload to send.
        :param num_repeats: The number of times we should send it.
        """

        if payload is None:
            payload = {}
        msg = msg_type(
            target_addr=self._mac_addr, source_id=self._source_id,
            seq_num=0, payload=payload,
            ack_requested=False, response_requested=False)
        self._loop.create_task(self._fire_sending(msg, num_repeats))

    async def _try_sending(
            self, msg: Message, response_type: Type[GenericResponse],
            *, timeout_secs: float=None, max_attempts: int=None) -> GenericResponse:
        """
        Send message and wait for appropriate response.

        :param msg: The message to be sent.
        :param timeout_secs: The timeout in seconds for each atempt.
        :param max_attempts: The maximum number of attempts.
        :return: The response we got.
        """
        assert self._transport is not None

        self._message[msg.seq_num] = [response_type, None, None]

        if timeout_secs is None:
            timeout_secs = self._timeout
        if max_attempts is None:
            max_attempts = self._retry_count

        attempts = 0
        while attempts < max_attempts:
            if msg.seq_num not in self._message:
                raise RuntimeError("Oops. We couldn't find the message information.")
            event = aio.Event()
            self._message[msg.seq_num][1] = event
            attempts += 1
            packed_message = msg.generate_packed_message()
            self._transport.sendto(packed_message)
            try:
                await aio.wait_for(event.wait(), timeout_secs)
                break
            except aio.TimeoutError:
                if attempts >= max_attempts:
                    if msg.seq_num in self._message:
                        del(self._message[msg.seq_num])
                    # It's dead Jim
                    self.unregister()
                    raise DeviceOffline()
        result = self._message[msg.seq_num][2]
        del (self._message[msg.seq_num])
        return cast(GenericResponse, result)

    async def _req_with_ack(
            self, msg_type: Type[Message], payload: Dict[str, Any],
            *, timeout_secs: int=None, max_attempts: int=None) -> msgtypes.Acknowledgement:
        """
        Send a message and expect an ACK response.

        :param msg_type: The type of the Message.
        :param payload: The payload to send.
        :param timeout_secs: The timeout in seconds for each atempt.
        :param max_attempts: The maximum number of attempts.
        :return:  The ACK response.

        Usually used for Set messages.
        """
        msg = msg_type(
            target_addr=self._mac_addr, source_id=self._source_id,
            seq_num=self._seq_next(),
            payload=payload, ack_requested=True, response_requested=False)
        return await self._try_sending(
            msg, msgtypes.Acknowledgement,
            timeout_secs=timeout_secs, max_attempts=max_attempts)

    # Usually used for Get messages, or for state confirmation after Set (hence the optional payload)
    async def _req_with_resp(
            self, msg_type: Type[Message], response_type: Type[GenericResponse],
            payload: Dict[str, Any]=None,
            *, timeout_secs: int=None, max_attempts: int=None) -> GenericResponse:
        """
        Send a message and expect an response.

        :param msg_type: The type of the Message.
        :param response_type: The type of the Response.
        :param payload: The payload to send.
        :param timeout_secs: The timeout in seconds for each atempt.
        :param max_attempts: The maximum number of attempts.
        :return:  The ACK response.

        Usually used for Get messages.
        """
        if payload is None:
            payload = {}
        msg = msg_type(
            target_addr=self._mac_addr, source_id=self._source_id,
            seq_num=self._seq_next(),
            payload=payload, ack_requested=False, response_requested=True)
        return await self._try_sending(
            msg, response_type, timeout_secs=timeout_secs, max_attempts=max_attempts)

    async def _req_with_ack_resp(
            self, msg_type: Type[Message], response_type: Type[GenericResponse],
            payload: Dict[str, str],
            *, timeout_secs: int=None, max_attempts: int=None) -> GenericResponse:
        """
        Send a message and expect an ACK and a response.

        :param msg_type: The type of the Message.
        :param response_type: The type of the Response.
        :param payload: The payload to send.
        :param timeout_secs: The timeout in seconds for each atempt.
        :param max_attempts: The maximum number of attempts.
        :return:  The ACK response.

        FIXME: Not currently implemented, although the LIFX LAN protocol supports
        this kind of workflow natively.
        """
        msg = msg_type(
            target_addr=self._mac_addr, source_id=self._source_id,
            seq_num=self._seq_next(),
            payload=payload, ack_requested=True, response_requested=True)
        return await self._try_sending(
            msg, response_type, timeout_secs=timeout_secs, max_attempts=max_attempts)

    #
    #                            Attribute Methods
    #
    async def get_label(self) -> str:
        """
        Get the label.

        :return: The current label.
        """
        label = self._label  # type: Optional[str]
        if label is None:
            resp = await self._req_with_resp(
                msgtypes.GetLabel, msgtypes.StateLabel)  # type: msgtypes.StateLabel
            self._label = resp.label.decode().replace("\x00", "")
        assert self._label is not None
        return self._label

    async def set_label(self, value: str) -> None:
        """
        Set the label.

        :param value: The new label.
        """
        if len(value) > 32:
            value = value[:32]
        await self._req_with_ack(msgtypes.SetLabel, {"label": value})
        self._label = value

    async def get_location(self) -> str:
        """
        Get the location.

        :return: The current location.
        """

        location = self._location  # type: Optional[str]
        if location is None:
            resp = await self._req_with_resp(
                msgtypes.GetLocation,
                msgtypes.StateLocation)  # type: msgtypes.StateLocation
            self._location = resp.label.decode().replace("\x00", "")
        assert self._location is not None
        return self._location

    # async def set_location(self, value):
    #     resp = await self.req_with_ack(SetLocation, {"location": value})
    #     self.resp_set_location(resp)
    #     return self.location

    async def get_group(self) -> str:
        """
        Get the group.

        :return: The current group
        """
        group = self._group  # type: Optional[str]
        if group is None:
            resp = await self._req_with_resp(
                msgtypes.GetGroup, msgtypes.StateGroup)  # type: msgtypes.StateGroup
            self._group = resp.label.decode().replace("\x00", "")
        assert self._group is not None
        return self._group

    # async def set_group(self, value):
    #     resp = await self.req_with_ack(SetGroup, {"group": value})
    #     self.resp_set_group(resp)
    #     return self.group

    async def get_power(self) -> Power:
        """
        Get the power setting.

        :return: The current power setting. Should normally be True or False.
        """
        table = {
            0: False,
            65535: True,
        }
        resp = await self._req_with_resp(
            msgtypes.GetPower, msgtypes.StatePower)  # type: msgtypes.StatePower
        self._power_level = table.get(resp.power_level, resp.power_level)
        assert self._power_level is not None
        return self._power_level

    async def set_power(self, value: Power, rapid: bool=False) -> None:
        """
        Set the current Power level.

        :param value: Normally True or False.
        :param rapid: If True then we don't wait for an ACK.
        """
        on = [True, 1, "on"]
        off = [False, 0, "off"]

        if value in on:
            value = 65535
        elif value in off:
            value = 0

        if not rapid:
            await self._req_with_ack(msgtypes.SetPower, {"power_level": value})
        else:
            self._fire_and_forget(msgtypes.SetPower, {"power_level": value})
        self._power_level = value

    async def get_wifi_firmware(self) -> Tuple[str, int]:
        """
        Get the WIFI Firmware version.

        :return: A tuple containing (version, timestamp).
        """
        wifi_firmware_version = self._wifi_firmware_version  # type: Optional[str]
        if wifi_firmware_version is None:
            resp = await self._req_with_resp(
                msgtypes.GetWifiFirmware,
                msgtypes.StateWifiFirmware)  # type: msgtypes.StateWifiFirmware
            self._wifi_firmware_version = (
                str(str(resp.version >> 16) + "." + str(resp.version & 0xff))
            )
            self._wifi_firmware_build_timestamp = resp.build
        assert self._wifi_firmware_version is not None
        assert self._wifi_firmware_build_timestamp is not None
        return self._wifi_firmware_version, self._wifi_firmware_build_timestamp

    async def get_wifi_info(self) -> msgtypes.StateWifiInfo:
        """
        Get the WIFI information.

        :return:The WIFI information.
        """
        return await self._req_with_resp(
            msgtypes.GetWifiInfo, msgtypes.StateWifiInfo)

    async def get_host_firmware(self) -> Tuple[str, int]:
        """
        Get the host Firmware version.

        :return: A tuple containing (version, timestamp).
        """
        host_firmware_version = self._host_firmware_version  # type: Optional[str]
        if host_firmware_version is None:
            resp = await self._req_with_resp(
                msgtypes.GetHostFirmware,
                msgtypes.StateHostFirmware)  # type: msgtypes.StateHostFirmware
            self._host_firmware_version = (
                str(str(resp.version >> 16) + "." + str(resp.version & 0xff))
            )
            self._host_firmware_build_timestamp = resp.build
        assert self._host_firmware_version is not None
        assert self._host_firmware_build_timestamp is not None
        return self._host_firmware_version, self._host_firmware_build_timestamp

    async def get_host_info(self) -> msgtypes.StateInfo:
        """
        Get the host information.

        :return: The host information.
        """
        return await self._req_with_resp(msgtypes.GetInfo, msgtypes.StateInfo)

    async def get_version(self) -> Tuple[int, int, int]:
        """
        Get the version information.

        :return: A tuple containing (vendor, product, version).
        """
        vendor = self._vendor  # type: Optional[int]
        if vendor is None:
            resp = await self._req_with_resp(
                msgtypes.GetVersion,
                msgtypes.StateVersion)  # type: msgtypes.StateVersion
            self._vendor = resp.vendor
            self._product = resp.product
            self._version = resp.version
        assert self._vendor is not None
        assert self._product is not None
        assert self._version is not None
        return self._vendor, self._product, self._version

    async def get_metadata(self, *, loop: aio.AbstractEventLoop) -> None:
        """
        Get and cache all meta data for device.

        :param loop: The asyncio event loop.

        This is a shortcut for running the following functions concurrently and
        waiting for all of them to return:

        * ``self.get_label()``
        * ``self.get_location()``
        * ``self.get_version()``
        * ``self.get_group()``
        * ``self.get_wifi_firmware()``
        * ``self.get_host_firmware()``
        """
        coroutines = [
            self.get_label(),
            self.get_location(),
            self.get_version(),
            self.get_group(),
            self.get_wifi_firmware(),
            self.get_host_firmware(),
        ]  # type: List[Awaitable]
        await aio.gather(*coroutines, loop=loop)

    #
    #                            Formating
    #
    def device_characteristics_str(self, indent: str) -> str:
        """
        Get a multi-line string with the device characteristics.

        :param indent: Prefix for each lines.
        :return: The resultant string.
        """
        s = "{}\n".format(self._label)
        s += indent + "MAC Address: {}\n".format(self._mac_addr)
        s += indent + "IP Address: {}\n".format(self._ip_addr)
        s += indent + "Port: {}\n".format(self._port)
        s += indent + "Power: {}\n".format(_str_map(self._power_level))
        s += indent + "Location: {}\n".format(self._location)
        s += indent + "Group: {}\n".format(self._group)
        return s

    def device_firmware_str(self, indent: str) -> str:
        """
        Get a multi-line string with the firmware information.

        :param indent: Prefix for each line.
        :return: The resultant string.
        """
        host_build_ns = self._host_firmware_build_timestamp
        if host_build_ns is not None:
            host_build_s = str(
                datetime.datetime.utcfromtimestamp(host_build_ns / 1000000000))
        else:
            host_build_s = "None"

        wifi_build_ns = self._wifi_firmware_build_timestamp
        if wifi_build_ns is not None:
            wifi_build_s = str(
                datetime.datetime.utcfromtimestamp(wifi_build_ns / 1000000000))
        else:
            wifi_build_s = "None"

        s = "Host Firmware Build Timestamp: {} ({} UTC)\n".format(host_build_ns, host_build_s)
        s += indent + "Host Firmware Build Version: {}\n".format(self._host_firmware_version)
        s += indent + "Wifi Firmware Build Timestamp: {} ({} UTC)\n".format(wifi_build_ns, wifi_build_s)
        s += indent + "Wifi Firmware Build Version: {}\n".format(self._wifi_firmware_version)
        return s

    def device_product_str(self, indent: str) -> str:
        """
        Get a multi-line string with the product information.

        :param indent: Prefix for each line.
        :return: The resultant string.
        """
        s = "Vendor: {}\n".format(self._vendor)
        s += indent + "Product: {}\n".format((self._product and product_map[self._product]) or "Unknown")
        s += indent + "Version: {}\n".format(self._version)
        return s

    @staticmethod
    def device_time_str(resp: msgtypes.StateInfo, indent: str="  ") -> str:
        """
        Get a multi-line string for the device information.

        :param resp: The device information.
        :param indent: Prefix for each line.
        :return: The resultant string.
        """
        dev_time = resp.time
        dev_uptime = resp.uptime
        dev_downtime = resp.downtime

        if dev_time is not None:
            time_s = str(datetime.datetime.utcfromtimestamp(dev_time / 1000000000))
        else:
            time_s = "None"

        if dev_uptime is not None:
            uptime_s = str(round(_nanosec_to_hours(dev_uptime), 2))
        else:
            uptime_s = "None"

        if dev_downtime is not None:
            downtime_s = str(round(_nanosec_to_hours(dev_downtime), 2))
        else:
            downtime_s = "None"

        s = "Current Time: {} ({} UTC)\n".format(dev_time, time_s)
        s += indent + "Uptime (ns): {} ({} hours)\n".format(dev_uptime, uptime_s)
        s += indent + "Last Downtime Duration +/-5s (ns): {} ({} hours)\n".format(
            dev_downtime, downtime_s)
        return s

    @staticmethod
    def device_radio_str(resp: msgtypes.StateWifiInfo, indent: str="  ") -> str:
        """
        Get a multi-line string for the wifi information.

        :param resp: The wifi information.
        :param indent: Prefix for each line.
        :return: The resultant string.
        """
        signal = resp.signal
        tx = resp.tx
        rx = resp.rx
        s = "Wifi Signal Strength (mW): {}\n".format(signal)
        s += indent + "Wifi TX (bytes): {}\n".format(tx)
        s += indent + "Wifi RX (bytes): {}\n".format(rx)
        return s

    def register_callback(self, call_back: Callable[[Message], None]) -> None:
        """
        Set callback to call when we receive an unexpected message.

        :param call_back: The call back function.
        """
        self._default_callback = call_back

    def __str__(self) -> str:
        """ Print identification. """
        return "{} ({})".format(self._label, self.mac_addr)

    def __repr__(self) -> str:
        """ Print identification. """
        return "<{} {} ({})>".format(type(self).__name__, self._label, self.mac_addr)


class Light(Device):
    """ Implement functions for a LIFX light. """

    def __init__(
            self, *, loop: aio.AbstractEventLoop,
            mac_addr: str, ip_addr: str,
            port: int=UDP_BROADCAST_PORT,
            devices: Devices=None) -> None:
        """
        Construct a new Light object.

        :param loop: The Asyncio event loop.
        :param mac_addr: The MAC Address. with the ":" and everything.
        :param ip_addr: A string with the IP address.
        :param port: The UDP port to use.
        :param devices: The lights list this light belongs to.
        """
        mac_addr = mac_addr.lower()
        super(Light, self).__init__(
            loop=loop, mac_addr=mac_addr, ip_addr=ip_addr, port=port, devices=devices)
        self._color = None  # type: Optional[Color]
        self._color_zones = []  # type: List[Color]
        self._infrared_brightness = None  # type: Optional[int]

    async def get_light_power(self) -> Power:
        """
        Get the light's power setting.

        :return: The light's power setting.
        """
        table = {
            0: False,
            65535: True,
        }
        resp = await self._req_with_resp(
            msgtypes.LightGetPower,
            msgtypes.LightStatePower)  # type: msgtypes.LightStatePower
        self._power_level = table.get(resp.power_level, resp.power_level)
        assert self._power_level is not None
        return self._power_level

    async def set_light_power(self, value: Power, duration: int=0, rapid: bool=False) -> None:
        """
        Ste the light's power setting.

        :param value: The new power setting.
        :param duration: The duration in ms to gradually make the change.
        :param rapid: If True then we don't wait for an ACK.
        """
        on = [True, 1, "on"]
        off = [False, 0, "off"]

        if value in on:
            value = 65535
        elif value in off:
            value = 0

        if not rapid:
            await self._req_with_ack(
                msgtypes.LightSetPower, {"power_level": value, "duration": duration})
        else:
            self._fire_and_forget(
                msgtypes.LightSetPower, {"power_level": value, "duration": duration},
                num_repeats=1)
        self._power_level = value

    async def get_color(self) -> Color:
        """
        Get the color of the light.

        :return: The color of the light.
        """
        resp = await self._req_with_resp(
            msgtypes.LightGet, msgtypes.LightState)  # type: msgtypes.LightState

        table = {
            0: False,
            65535: True,
        }
        self._power_level = table.get(resp.power_level, resp.power_level)
        self._color = Color.create_from_values(resp.color)
        self._label = resp.label.decode().replace("\x00", "")

        return self._color

    async def set_color(self, color: Color, duration: int=0, rapid: bool=False) -> None:
        """
        Set the color of the light.

        :param color: Input colour.
        :param duration: Time to make change in ms.
        :param rapid: If True then we don't wait for an ACK.
        :return: The new color of the light.
        """
        value = color.get_values()
        if rapid:
            self._fire_and_forget(
                msgtypes.LightSetColor,
                {"color": value, "duration": duration},
                num_repeats=1)
        else:
            await self._req_with_ack(
                msgtypes.LightSetColor,
                {"color": value, "duration": duration})
        self._color = color

    async def get_color_zones(self, start_index: int, end_index: int=None) -> List[Color]:
        """
        Get color zones.

        :param start_index: The start index.
        :param end_index: The end Index.
        """
        if end_index is None:
            end_index = start_index + 8
        args = {
            "start_index": start_index,
            "end_index": end_index,
        }
        resp = await self._req_with_resp(
            msgtypes.MultiZoneGetColorZones,
            msgtypes.MultiZoneStateMultiZone,
            payload=args)  # type: msgtypes.MultiZoneStateMultiZone

        self._color_zones = []
        for HSBK in resp.color:   # type: Tuple[int, int, int, int]
            self._color_zones.append(Color.create_from_values(HSBK))

        return self._color_zones

    async def set_color_zones(
            self, start_index: int, end_index: int, color: Color,
            duration: int=0, apply: int=1, rapid: bool=False) -> None:
        """
        Set color zones.

        :param start_index:  The start index.
        :param end_index: The end index,
        :param color: The colour.
        :param duration: The duration in ms.
        :param apply: The apply value.
        :param rapid: If True then we don't wait for an ACK.
        """
        args = {
            "start_index": start_index,
            "end_index": end_index,
            "color": color.get_values(),
            "duration": duration,
            "apply": apply,
        }

        if rapid:
            self._fire_and_forget(
                msgtypes.MultiZoneSetColorZones, args, num_repeats=1)
        else:
            await self._req_with_ack(
                msgtypes.MultiZoneSetColorZones, args)

    async def set_waveform(
            self, *,
            color: Color,
            transient: int, period: int, cycles: int, duty_cycle: int, waveform: int,
            rapid: bool=False) -> None:
        """
        Set the Wave form.

        :param color: The color used.
        :param transient: The Transient value.
        :param period: The period in ms.
        :param cycles: The number of cycles.
        :param duty_cycle: The number of duty cycles.
        :param waveform: The waveform value.
        :param rapid: If True then we don't wait for an ACK.
        """
        value = {
            'color': color.get_values(),
            'transient': transient,
            'period': period,
            'cycles': cycles,
            'duty_cycle': duty_cycle,
            'waveform': waveform,
        }

        if rapid:
            self._fire_and_forget(
                msgtypes.LightSetWaveform, value, num_repeats=1)
        else:
            await self._req_with_ack(
                msgtypes.LightSetWaveform, value)

    async def get_infrared(self) -> int:
        """
        Get infra-red brightness.
        :return: Number 0-100.
        """
        resp = await self._req_with_resp(
            msgtypes.LightGetInfrared,
            msgtypes.LightStateInfrared)  # type: msgtypes.LightStateInfrared
        self._infrared_brightness = int(resp.infrared_brightness * 100 / 65535)
        return self._infrared_brightness

    async def set_infrared(self, infrared_brightness: int, rapid: bool=False) -> None:
        """
        Set infra-red brightness.

        :param infrared_brightness:  Number 0-100.
        :param rapid: If True then we don't wait for an ACK.
        """
        value = int(infrared_brightness * 65535 / 100)
        if rapid:
            self._fire_and_forget(
                msgtypes.LightSetInfrared, {"infrared_brightness": value}, num_repeats=1)
        else:
            await self._req_with_ack(
                msgtypes.LightSetInfrared, {"infrared_brightness": value})
        self._infrared_brightness = value


class LifxDiscovery(aio.DatagramProtocol):
    """ A protocol handler that discovers Lifx Lights. """

    def __init__(
            self, *,
            loop: aio.AbstractEventLoop,
            devices: Devices=None,
            ipv6prefix: str=None,
            discovery_interval: int=DISCOVERY_INTERVAL,
            discovery_step: int=DISCOVERY_STEP) -> None:
        """
        Construct an `LifxDiscovery` object.

        :param loop: The asyncio event loop.
        :param devices: The devices object to contain lights.
        :param ipv6prefix: The IPv6 prefix to use for IPv6 addresses.
        :param discovery_interval: How often should we rerun discover (seconds)?
        :param discovery_step: How often should we wake up (seconds)?
        """
        self._lights = {}  # type: Dict[str, Light]
        self._devices = devices  # Where to register new devices
        self._transport = None  # type: Optional[aio.DatagramTransport]
        self._loop = loop
        self._source_id = random.randint(0, (2 ** 32) - 1)
        self._ipv6prefix = ipv6prefix
        self._discovery_interval = discovery_interval
        self._discovery_step = discovery_step
        self._discovery_countdown = 0

    def connection_made(self, transport: aio.BaseTransport) -> None:
        """ Called when we receive a connection. """
        self._transport = cast(aio.DatagramTransport, transport)
        sock = self._transport.get_extra_info("socket")
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        self._loop.call_soon(self._discover)

    def datagram_received(self, data: AnyStr, addr: str) -> None:
        """ Called when we receive a packet. """
        assert isinstance(data, bytes)
        response = unpack_lifx_message(data)
        ip_addr = addr[0]

        mac_addr = response.target_addr
        if mac_addr == BROADCAST_MAC:
            return

        if type(response) == msgtypes.StateService:
            # discovered
            assert isinstance(response, msgtypes.StateService)
            if response.service == 1:  # only look for UDP services
                remote_port = response.port
            else:
                return
        elif type(response) == msgtypes.LightState:
            # looks like the lights are volunteering LightState after booting
            remote_port = UDP_BROADCAST_PORT
        else:
            return

        if self._ipv6prefix:
            family = socket.AF_INET6
            remote_ip = _mac_to_ipv6_link_local(mac_addr, self._ipv6prefix)
        else:
            family = socket.AF_INET
            remote_ip = ip_addr

        if mac_addr in self._lights:
            # rediscovered
            light = self._lights[mac_addr]  # type: Light
        else:
            # newly discovered
            light = Light(
                loop=self._loop,
                mac_addr=mac_addr,
                ip_addr=remote_ip,
                port=remote_port,
                devices=self._devices)
            self._lights[mac_addr] = light
        light.renew(family=family, ip_addr=remote_ip, port=remote_port)

    def _discover(self) -> None:
        """ Called regularly based on ``discovery_step`` parameter. """

        if self._transport:
            assert self._transport is not None

            if self._discovery_countdown <= 0:
                self._discovery_countdown = self._discovery_interval
                msg = msgtypes.GetService(
                    target_addr=BROADCAST_MAC, source_id=self._source_id,
                    seq_num=0, payload={},
                    ack_requested=False, response_requested=True)
                self._transport.sendto(msg.generate_packed_message(), (UDP_BROADCAST_IP, UDP_BROADCAST_PORT))
            else:
                self._discovery_countdown -= self._discovery_step
            self._loop.call_later(self._discovery_step, self._discover)

    def _register(self, alight: Light) -> None:
        """ Register discovered light. """
        if self._devices:
            self._devices.register(alight)

    def _unregister(self, alight: Light) -> None:
        """ Unregister lost light. """
        if self._devices:
            self._devices.unregister(alight)

    def _cleanup(self) -> None:
        """ Cleanup. FIXME: Is the actually used??? """
        if self._transport:
            self._transport.close()
            self._transport = None
        for light in self._lights.values():
            light.cleanup()
        self._lights = {}



product_map = {1: "Original 1000",
               3: "Color 650",
               10: "White 800 (Low Voltage)",
               11: "White 800 (High Voltage)",
               18: "White 900 BR30 (Low Voltage)",
               20: "Color 1000 BR30",
               22: "Color 1000",
               27: "LIFX A19",
               28: "LIFX BR30",
               29: "LIFX+ A19",
               30: "LIFX+ BR30",
               31: "LIFX Z"
               }

features_map = {1: {"color": True,
                    "infrared": False,
                    "multizone": False},
                3: {"color": True,
                    "infrared": False,
                    "multizone": False},
                10: {"color": False,
                     "infrared": False,
                     "multizone": False},
                11: {"color": False,
                     "infrared": False,
                     "multizone": False},
                18: {"color": False,
                     "infrared": False,
                     "multizone": False},
                20: {"color": True,
                     "infrared": False,
                     "multizone": False},
                22: {"color": True,
                     "infrared": False,
                     "multizone": False},
                27: {"color": True,
                     "infrared": False,
                     "multizone": False},
                28: {"color": True,
                     "infrared": False,
                     "multizone": False},
                29: {"color": True,
                     "infrared": True,
                     "multizone": False},
                30: {"color": True,
                     "infrared": True,
                     "multizone": False},
                31: {"color": True,
                     "infrared": False,
                     "multizone": True}
                }



# unpack.py
# Author: Meghan Clark
import struct

from .message import Message, HEADER_SIZE_BYTES
from . import msgtypes
import binascii

# Creates a LIFX Message out of packed binary data
# If the message type is not one of the officially released ones above, it will create just a Message out of it
# If it's not in the LIFX protocol format, uhhhhh...we'll put that on a to-do list.


def unpack_lifx_message(packed_message: bytes) -> Message:
    header_str = packed_message[0:HEADER_SIZE_BYTES]
    payload_str = packed_message[HEADER_SIZE_BYTES:]

    size = struct.unpack("H", header_str[0:2])[0]
    flags = struct.unpack("H", header_str[2:4])[0]
    origin = (flags >> 14) & 3
    tagged = (flags >> 13) & 1
    addressable = (flags >> 12) & 1
    protocol = flags & 4095
    source_id = struct.unpack("I", header_str[4:8])[0]
    target_addr = ":".join([('%02x' % b) for b in struct.unpack("B" * 6, header_str[8:14])])
    response_flags = struct.unpack("B", header_str[22:23])[0]
    ack_requested = bool(response_flags & 2)
    response_requested = bool(response_flags & 1)
    seq_num = struct.unpack("B", header_str[23:24])[0]
    message_type = struct.unpack("H", header_str[32:34])[0]

    message = Message(
        target_addr=target_addr, source_id=source_id, seq_num=seq_num,
        payload={},
        ack_requested=ack_requested, response_requested=response_requested)
    if message_type == msgtypes.MSG_IDS[msgtypes.GetService]:
        message = msgtypes.GetService(
            target_addr=target_addr, source_id=source_id, seq_num=seq_num,
            payload={},
            ack_requested=ack_requested, response_requested=response_requested)

    elif message_type == msgtypes.MSG_IDS[msgtypes.StateService]:
        service = struct.unpack("B", payload_str[0:1])[0]
        port = struct.unpack("I", payload_str[1:5])[0]
        payload = {"service": service, "port": port}
        message = msgtypes.StateService(
            target_addr=target_addr, source_id=source_id, seq_num=seq_num,
            payload=payload,
            ack_requested=ack_requested, response_requested=response_requested)

    elif message_type == msgtypes.MSG_IDS[msgtypes.GetHostInfo]:
        message = msgtypes.GetHostInfo(
            target_addr=target_addr, source_id=source_id, seq_num=seq_num,
            payload={},
            ack_requested=ack_requested, response_requested=response_requested)

    elif message_type == msgtypes.MSG_IDS[msgtypes.StateHostInfo]:
        signal = struct.unpack("f", payload_str[0:4])[0]
        tx = struct.unpack("I", payload_str[4:8])[0]
        rx = struct.unpack("I", payload_str[8:12])[0]
        reserved1 = struct.unpack("h", payload_str[12:14])[0]
        payload = {"signal": signal, "tx": tx, "rx": rx, "reserved1": reserved1}
        message = msgtypes.StateHostInfo(
            target_addr=target_addr, source_id=source_id, seq_num=seq_num,
            payload=payload,
            ack_requested=ack_requested, response_requested=response_requested)

    elif message_type == msgtypes.MSG_IDS[msgtypes.GetHostFirmware]:
        message = msgtypes.GetHostFirmware(
            target_addr=target_addr, source_id=source_id, seq_num=seq_num,
            payload={},
            ack_requested=ack_requested, response_requested=response_requested)

    elif message_type == msgtypes.MSG_IDS[msgtypes.StateHostFirmware]:
        build = struct.unpack("Q", payload_str[0:8])[0]
        reserved1 = struct.unpack("Q", payload_str[8:16])[0]
        version = struct.unpack("I", payload_str[16:20])[0]
        payload = {"build": build, "reserved1": reserved1, "version": version}
        message = msgtypes.StateHostFirmware(
            target_addr=target_addr, source_id=source_id, seq_num=seq_num,
            payload=payload,
            ack_requested=ack_requested, response_requested=response_requested)

    elif message_type == msgtypes.MSG_IDS[msgtypes.GetWifiInfo]:
        message = msgtypes.GetWifiInfo(
            target_addr=target_addr, source_id=source_id, seq_num=seq_num,
            payload={},
            ack_requested=ack_requested, response_requested=response_requested)

    elif message_type == msgtypes.MSG_IDS[msgtypes.StateWifiInfo]:
        signal = struct.unpack("f", payload_str[0:4])[0]
        tx = struct.unpack("I", payload_str[4:8])[0]
        rx = struct.unpack("I", payload_str[8:12])[0]
        reserved1 = struct.unpack("h", payload_str[12:14])[0]
        payload = {"signal": signal, "tx": tx, "rx": rx, "reserved1": reserved1}
        message = msgtypes.StateWifiInfo(
            target_addr=target_addr, source_id=source_id, seq_num=seq_num,
            payload=payload,
            ack_requested=ack_requested, response_requested=response_requested)

    elif message_type == msgtypes.MSG_IDS[msgtypes.GetWifiFirmware]:
        message = msgtypes.GetWifiFirmware(
            target_addr=target_addr, source_id=source_id, seq_num=seq_num,
            payload={},
            ack_requested=ack_requested, response_requested=response_requested)

    elif message_type == msgtypes.MSG_IDS[msgtypes.StateWifiFirmware]:
        build = struct.unpack("Q", payload_str[0:8])[0]
        reserved1 = struct.unpack("Q", payload_str[8:16])[0]
        version = struct.unpack("I", payload_str[16:20])[0]
        payload = {"build": build, "reserved1": reserved1, "version": version}
        message = msgtypes.StateWifiFirmware(
            target_addr=target_addr, source_id=source_id, seq_num=seq_num,
            payload=payload,
            ack_requested=ack_requested, response_requested=response_requested)

    elif message_type == msgtypes.MSG_IDS[msgtypes.GetPower]:
        message = msgtypes.GetPower(
            target_addr=target_addr, source_id=source_id, seq_num=seq_num,
            payload={},
            ack_requested=ack_requested, response_requested=response_requested)

    elif message_type == msgtypes.MSG_IDS[msgtypes.SetPower]:
        power_level = struct.unpack("H", payload_str[0:2])[0]
        payload = {"power_level": power_level}
        message = msgtypes.SetPower(
            target_addr=target_addr, source_id=source_id, seq_num=seq_num,
            payload=payload,
            ack_requested=ack_requested, response_requested=response_requested)

    elif message_type == msgtypes.MSG_IDS[msgtypes.StatePower]:
        power_level = struct.unpack("H", payload_str[0:2])[0]
        payload = {"power_level": power_level}
        message = msgtypes.StatePower(
            target_addr=target_addr, source_id=source_id, seq_num=seq_num,
            payload=payload,
            ack_requested=ack_requested, response_requested=response_requested)

    elif message_type == msgtypes.MSG_IDS[msgtypes.GetLabel]:
        message = msgtypes.GetLabel(
            target_addr=target_addr, source_id=source_id, seq_num=seq_num,
            payload={},
            ack_requested=ack_requested, response_requested=response_requested)

    elif message_type == msgtypes.MSG_IDS[msgtypes.SetLabel]:
        label = binascii.unhexlify("".join(["%2.2x" % (b & 0x000000ff)
                                            for b in struct.unpack("b" * 32, payload_str[0:32])]))
        payload = {"label": label}
        message = msgtypes.SetLabel(
            target_addr=target_addr, source_id=source_id, seq_num=seq_num,
            payload=payload,
            ack_requested=ack_requested, response_requested=response_requested)

    elif message_type == msgtypes.MSG_IDS[msgtypes.StateLabel]:
        label = binascii.unhexlify("".join(["%2.2x" % (b & 0x000000ff)
                                            for b in struct.unpack("b" * 32, payload_str[0:32])]))
        payload = {"label": label}
        message = msgtypes.StateLabel(
            target_addr=target_addr, source_id=source_id, seq_num=seq_num,
            payload=payload,
            ack_requested=ack_requested, response_requested=response_requested)

    elif message_type == msgtypes.MSG_IDS[msgtypes.GetLocation]:
        message = msgtypes.GetLocation(
            target_addr=target_addr, source_id=source_id, seq_num=seq_num,
            payload={},
            ack_requested=ack_requested, response_requested=response_requested)

    elif message_type == msgtypes.MSG_IDS[msgtypes.StateLocation]:
        location = [b for b in struct.unpack("B" * 16, payload_str[0:16])]
        label = binascii.unhexlify("".join(["%2.2x" % (b & 0x000000ff)
                                            for b in struct.unpack("b" * 32, payload_str[16:48])]))
        updated_at = struct.unpack("Q", payload_str[48:56])[0]
        payload = {"location": location, "label": label, "updated_at": updated_at}
        message = msgtypes.StateLocation(
            target_addr=target_addr, source_id=source_id, seq_num=seq_num,
            payload=payload,
            ack_requested=ack_requested, response_requested=response_requested)

    elif message_type == msgtypes.MSG_IDS[msgtypes.GetGroup]:
        message = msgtypes.GetGroup(
            target_addr=target_addr, source_id=source_id, seq_num=seq_num,
            payload={},
            ack_requested=ack_requested, response_requested=response_requested)

    elif message_type == msgtypes.MSG_IDS[msgtypes.StateGroup]:
        group = [b for b in struct.unpack("B" * 16, payload_str[0:16])]
        label = binascii.unhexlify("".join(["%2.2x" % (b & 0x000000ff)
                                            for b in struct.unpack("b" * 32, payload_str[16:48])]))
        updated_at = struct.unpack("Q", payload_str[48:56])[0]
        payload = {"group": group, "label": label, "updated_at": updated_at}
        message = msgtypes.StateGroup(
            target_addr=target_addr, source_id=source_id, seq_num=seq_num,
            payload=payload,
            ack_requested=ack_requested, response_requested=response_requested)

    elif message_type == msgtypes.MSG_IDS[msgtypes.GetVersion]:
        message = msgtypes.GetVersion(
            target_addr=target_addr, source_id=source_id, seq_num=seq_num,
            payload={},
            ack_requested=ack_requested, response_requested=response_requested)

    elif message_type == msgtypes.MSG_IDS[msgtypes.StateVersion]:
        vendor = struct.unpack("I", payload_str[0:4])[0]
        product = struct.unpack("I", payload_str[4:8])[0]
        version = struct.unpack("I", payload_str[8:12])[0]
        payload = {"vendor": vendor, "product": product, "version": version}
        message = msgtypes.StateVersion(
            target_addr=target_addr, source_id=source_id, seq_num=seq_num,
            payload=payload,
            ack_requested=ack_requested, response_requested=response_requested)

    elif message_type == msgtypes.MSG_IDS[msgtypes.GetInfo]:
        message = msgtypes.GetInfo(
            target_addr=target_addr, source_id=source_id, seq_num=seq_num,
            payload={},
            ack_requested=ack_requested, response_requested=response_requested)

    elif message_type == msgtypes.MSG_IDS[msgtypes.StateInfo]:
        time = struct.unpack("Q", payload_str[0:8])[0]
        uptime = struct.unpack("Q", payload_str[8:16])[0]
        downtime = struct.unpack("Q", payload_str[16:24])[0]
        payload = {"time": time, "uptime": uptime, "downtime": downtime}
        message = msgtypes.StateInfo(
            target_addr=target_addr, source_id=source_id, seq_num=seq_num,
            payload=payload,
            ack_requested=ack_requested, response_requested=response_requested)

    elif message_type == msgtypes.MSG_IDS[msgtypes.Acknowledgement]:
        message = msgtypes.Acknowledgement(
            target_addr=target_addr, source_id=source_id, seq_num=seq_num,
            payload={},
            ack_requested=ack_requested, response_requested=response_requested)

    elif message_type == msgtypes.MSG_IDS[msgtypes.EchoRequest]:
        byte_array_len = len(payload_str)
        byte_array = [b for b in struct.unpack("B" * byte_array_len, payload_str[0:byte_array_len])]
        payload = {"byte_array": byte_array}
        message = msgtypes.EchoRequest(
            target_addr=target_addr, source_id=source_id, seq_num=seq_num,
            payload=payload,
            ack_requested=ack_requested, response_requested=response_requested)

    elif message_type == msgtypes.MSG_IDS[msgtypes.EchoResponse]:
        byte_array_len = len(payload_str)
        byte_array = [b for b in struct.unpack("B" * byte_array_len, payload_str[0:byte_array_len])]
        payload = {"byte_array": byte_array}
        message = msgtypes.EchoResponse(
            target_addr=target_addr, source_id=source_id, seq_num=seq_num,
            payload=payload,
            ack_requested=ack_requested, response_requested=response_requested)

    elif message_type == msgtypes.MSG_IDS[msgtypes.LightGet]:
        message = msgtypes.LightGet(
            target_addr=target_addr, source_id=source_id, seq_num=seq_num,
            payload={},
            ack_requested=ack_requested, response_requested=response_requested)

    elif message_type == msgtypes.MSG_IDS[msgtypes.LightSetColor]:
        color = struct.unpack("H" * 4, payload_str[0:8])
        duration = struct.unpack("I", payload_str[8:12])[0]
        payload = {"color": color, "duration": duration}
        message = msgtypes.LightSetColor(
            target_addr=target_addr, source_id=source_id, seq_num=seq_num,
            payload=payload,
            ack_requested=ack_requested, response_requested=response_requested)

    elif message_type == msgtypes.MSG_IDS[msgtypes.LightState]:
        color = struct.unpack("H" * 4, payload_str[0:8])
        reserved1 = struct.unpack("H", payload_str[8:10])[0]
        power_level = struct.unpack("H", payload_str[10:12])[0]
        label = binascii.unhexlify("".join(["%2.2x" % (b & 0x000000ff)
                                            for b in struct.unpack("b" * 32, payload_str[12:44])]))
        reserved2 = struct.unpack("Q", payload_str[44:52])[0]
        payload = {"color": color, "reserved1": reserved1,
                   "power_level": power_level, "label": label, "reserved2": reserved2}
        message = msgtypes.LightState(
            target_addr=target_addr, source_id=source_id, seq_num=seq_num,
            payload=payload,
            ack_requested=ack_requested, response_requested=response_requested)

    elif message_type == msgtypes.MSG_IDS[msgtypes.LightGetPower]:
        message = msgtypes.LightGetPower(
            target_addr=target_addr, source_id=source_id, seq_num=seq_num,
            payload={},
            ack_requested=ack_requested, response_requested=response_requested)

    elif message_type == msgtypes.MSG_IDS[msgtypes.LightSetPower]:
        power_level = struct.unpack("H", payload_str[0:2])[0]
        duration = struct.unpack("I", payload_str[2:6])[0]
        payload = {"power_level": power_level, "duration": duration}
        message = msgtypes.LightSetPower(
            target_addr=target_addr, source_id=source_id, seq_num=seq_num,
            payload=payload,
            ack_requested=ack_requested, response_requested=response_requested)

    elif message_type == msgtypes.MSG_IDS[msgtypes.LightStatePower]:
        power_level = struct.unpack("H", payload_str[0:2])[0]
        payload = {"power_level": power_level}
        message = msgtypes.LightStatePower(
            target_addr=target_addr, source_id=source_id, seq_num=seq_num,
            payload=payload,
            ack_requested=ack_requested, response_requested=response_requested)

    elif message_type == msgtypes.MSG_IDS[msgtypes.LightGetInfrared]:  # 120
        message = msgtypes.LightGetInfrared(
            target_addr=target_addr, source_id=source_id, seq_num=seq_num,
            payload={},
            ack_requested=ack_requested, response_requested=response_requested)

    elif message_type == msgtypes.MSG_IDS[msgtypes.LightStateInfrared]:  # 121
        infrared_brightness = struct.unpack("H", payload_str[0:2])[0]
        payload = {"infrared_brightness": infrared_brightness}
        message = msgtypes.LightStateInfrared(
            target_addr=target_addr, source_id=source_id, seq_num=seq_num,
            payload=payload,
            ack_requested=ack_requested, response_requested=response_requested)

    elif message_type == msgtypes.MSG_IDS[msgtypes.LightSetInfrared]:  # 122
        infrared_brightness = struct.unpack("H", payload_str[0:2])[0]
        payload = {"infrared_brightness": infrared_brightness}
        message = msgtypes.LightSetInfrared(
            target_addr=target_addr, source_id=source_id, seq_num=seq_num,
            payload=payload,
            ack_requested=ack_requested, response_requested=response_requested)

    elif message_type == msgtypes.MSG_IDS[msgtypes.MultiZoneStateZone]:  # 503
        count = struct.unpack("c", payload_str[0:1])[0]
        count = ord(count)  # 8 bit
        index = struct.unpack("c", payload_str[1:2])[0]
        index = ord(index)  # 8 bit
        color = struct.unpack("H" * 4, payload_str[2:10])
        payload = {"count": count, "index": index, "color": color}
        message = msgtypes.MultiZoneStateZone(
            target_addr=target_addr, source_id=source_id, seq_num=seq_num,
            payload=payload,
            ack_requested=ack_requested, response_requested=response_requested)

    elif message_type == msgtypes.MSG_IDS[msgtypes.MultiZoneStateMultiZone]:  # 506
        count = struct.unpack("c", payload_str[0:1])[0]
        count = ord(count)  # 8 bit
        index = struct.unpack("c", payload_str[1:2])[0]
        index = ord(index)  # 8 bit
        colors = []
        for i in range(8):
            color = struct.unpack("H" * 4, payload_str[2 + (i * 8):10 + (i * 8)])
            colors.append(color)
        payload = {"count": count, "index": index, "color": colors}
        message = msgtypes.MultiZoneStateMultiZone(
            target_addr=target_addr, source_id=source_id, seq_num=seq_num,
            payload=payload,
            ack_requested=ack_requested, response_requested=response_requested)

    else:
        message = Message(
            target_addr=target_addr, source_id=source_id, seq_num=seq_num,
            payload={},
            ack_requested=ack_requested, response_requested=response_requested)
        message.message_type = message_type

    message.size = size
    message.origin = origin
    message.tagged = tagged
    message.addressable = addressable
    message.protocol = protocol
    message.source_id = source_id
    message.header = header_str
    message.payload = payload_str

    return message



""" Define colours for use with aiolifxc. """
from typing import Tuple

UINT16_MAX = pow(2, 16) - 1

HUE_MAX = 360
KELVIN_MIN = 2500
KELVIN_MAX = 9000
KELVIN_RANGE = KELVIN_MAX - KELVIN_MIN

MID_KELVIN = int(KELVIN_MIN + (KELVIN_RANGE/2))


class Color:
    """ An immutable type representing a colour. """
    def __init__(self, hue: int, saturation: int, brightness: int, kelvin: int) -> None:
        """
        Create a new colour using HSBK.

        :param hue: Range 0 to 360.
        :param saturation: Range 0 to 100.
        :param brightness: Range 0 to 100.
        :param kelvin: Range 2500 to 9000.
        """
        self._hue = hue
        self._saturation = saturation
        self._brightness = brightness
        self._kelvin = kelvin

    def clone(self) -> 'Color':
        return type(self)(self._hue, self._saturation, self._brightness, self._kelvin)

    def get_values(self) -> Tuple[int, int, int, int]:
        return (
            int(self._hue / HUE_MAX * UINT16_MAX),
            int(self._saturation / 100 * UINT16_MAX),
            int(self._brightness / 100 * UINT16_MAX),
            self._kelvin,
        )

    @classmethod
    def create_from_values(cls, values: Tuple[int, int, int, int]) -> 'Color':
        return cls(
            hue=int(values[0] / UINT16_MAX * HUE_MAX),
            saturation=int(values[1] / UINT16_MAX * 100),
            brightness=int(values[2] / UINT16_MAX * 100),
            kelvin=values[3],
        )

    def __str__(self) -> str:
        return "Colour: %d %d %d %d" % (
            self._hue, self._saturation, self._brightness, self._kelvin)


# Bright Colors
RED = Color(0, 100, 100, MID_KELVIN)
YELLOW = Color(60, 100, 100, MID_KELVIN)
GREEN = Color(120, 100, 100, MID_KELVIN)
AQUA = Color(180, 100, 100, MID_KELVIN)
BLUE = Color(240, 100, 100, MID_KELVIN)
PURPLE = Color(300, 100, 100, MID_KELVIN)
WHITE = Color(0, 0, 100, MID_KELVIN)

# Whites
COOL_WHITE = Color(0, 0, 100, KELVIN_MAX)
WARM_WHITE = Color(0, 0, 100, KELVIN_MIN)



# msgtypes.py
# Author: Meghan Clark
#    Edited for python3 by François Wautier

# To Do: Validate that args are within required ranges, types, etc. In
# particular: Color [0-65535, 0-65535, 0-65535, 2500-9000], Power Level (must be 0 OR 65535)
# Need to look into assert-type frameworks or something, there has to be a tool for that.
# Also need to make custom errors possibly, though tool may have those.
from typing import Dict, Any

from .message import Message, little_endian
import bitstring

# DEVICE MESSAGES


class GetService(Message):
    def __init__(
            self, *, target_addr: str, source_id: int, seq_num: int,
            payload: Dict[str, Any]=None,
            ack_requested: bool=False, response_requested: bool=False) -> None:

        if payload is None:
            payload = {}

        super().__init__(
            target_addr=target_addr, source_id=source_id,
            seq_num=seq_num,
            ack_requested=ack_requested, response_requested=response_requested,
            payload=payload)
        self.message_type = MSG_IDS[GetService]


class StateService(Message):
    def __init__(
            self, *, target_addr: str, source_id: int, seq_num: int,
            payload: Dict[str, Any]=None,
            ack_requested: bool=False, response_requested: bool=False) -> None:

        if payload is None:
            payload = {}

        self.service = payload["service"]
        self.port = payload["port"]
        super().__init__(
            target_addr=target_addr, source_id=source_id,
            seq_num=seq_num,
            ack_requested=ack_requested, response_requested=response_requested,
            payload=payload)
        self.message_type = MSG_IDS[StateService]

    def get_payload(self) -> bytes:
        self.payload_fields.append(("Service", self.service))
        self.payload_fields.append(("Port", self.port))
        service = little_endian(bitstring.pack("uint:8", self.service))
        port = little_endian(bitstring.pack("uint:32", self.port))
        payload = service + port
        return payload


class GetHostInfo(Message):
    def __init__(
            self, *, target_addr: str, source_id: int, seq_num: int,
            payload: Dict[str, Any]=None,
            ack_requested: bool=False, response_requested: bool=False) -> None:

        if payload is None:
            payload = {}

        super().__init__(
            target_addr=target_addr, source_id=source_id,
            seq_num=seq_num,
            ack_requested=ack_requested, response_requested=response_requested,
            payload=payload)
        self.message_type = MSG_IDS[GetHostInfo]


class StateHostInfo(Message):
    def __init__(
            self, *, target_addr: str, source_id: int, seq_num: int,
            payload: Dict[str, Any]=None,
            ack_requested: bool=False, response_requested: bool=False) -> None:

        if payload is None:
            payload = {}

        self.signal = payload["signal"]
        self.tx = payload["tx"]
        self.rx = payload["rx"]
        self.reserved1 = payload["reserved1"]
        super().__init__(
            target_addr=target_addr, source_id=source_id,
            seq_num=seq_num,
            ack_requested=ack_requested, response_requested=response_requested,
            payload=payload)
        self.message_type = MSG_IDS[StateHostInfo]

    def get_payload(self) -> bytes:
        self.payload_fields.append(("Signal (mW)", self.signal))
        self.payload_fields.append(("TX (bytes since on)", self.tx))
        self.payload_fields.append(("RX (bytes since on)", self.rx))
        self.payload_fields.append(("Reserved", self.reserved1))
        signal = little_endian(bitstring.pack("float:32", self.signal))
        tx = little_endian(bitstring.pack("uint:32", self.tx))
        rx = little_endian(bitstring.pack("uint:32", self.rx))
        reserved1 = little_endian(bitstring.pack("int:16", self.reserved1))
        payload = signal + tx + rx + reserved1
        return payload


class GetHostFirmware(Message):
    def __init__(
            self, *, target_addr: str, source_id: int, seq_num: int,
            payload: Dict[str, Any]=None,
            ack_requested: bool=False, response_requested: bool=False) -> None:

        if payload is None:
            payload = {}

        super().__init__(
            target_addr=target_addr, source_id=source_id,
            seq_num=seq_num,
            ack_requested=ack_requested, response_requested=response_requested,
            payload=payload)
        self.message_type = MSG_IDS[GetHostFirmware]


class StateHostFirmware(Message):
    def __init__(
            self, *, target_addr: str, source_id: int, seq_num: int,
            payload: Dict[str, Any]=None,
            ack_requested: bool=False, response_requested: bool=False) -> None:

        if payload is None:
            payload = {}

        self.build = payload["build"]
        self.reserved1 = payload["reserved1"]
        self.version = payload["version"]
        super().__init__(
            target_addr=target_addr, source_id=source_id,
            seq_num=seq_num,
            ack_requested=ack_requested, response_requested=response_requested,
            payload=payload)
        self.message_type = MSG_IDS[StateHostFirmware]

    def get_payload(self) -> bytes:
        self.payload_fields.append(("Timestamp of Build", self.build))
        self.payload_fields.append(("Reserved", self.reserved1))
        self.payload_fields.append(("Version", self.version))
        build = little_endian(bitstring.pack("uint:64", self.build))
        reserved1 = little_endian(bitstring.pack("uint:64", self.reserved1))
        version = little_endian(bitstring.pack("uint:32", self.version))
        payload = build + reserved1 + version
        return payload


class GetWifiInfo(Message):
    def __init__(
            self, *, target_addr: str, source_id: int, seq_num: int,
            payload: Dict[str, Any]=None,
            ack_requested: bool=False, response_requested: bool=False) -> None:

        if payload is None:
            payload = {}

        super().__init__(
            target_addr=target_addr, source_id=source_id,
            seq_num=seq_num,
            ack_requested=ack_requested, response_requested=response_requested,
            payload=payload)
        self.message_type = MSG_IDS[GetWifiInfo]


class StateWifiInfo(Message):
    def __init__(
            self, *, target_addr: str, source_id: int, seq_num: int,
            payload: Dict[str, Any]=None,
            ack_requested: bool=False, response_requested: bool=False) -> None:

        if payload is None:
            payload = {}

        self.signal = payload["signal"]
        self.tx = payload["tx"]
        self.rx = payload["rx"]
        self.reserved1 = payload["reserved1"]
        super().__init__(
            target_addr=target_addr, source_id=source_id,
            seq_num=seq_num,
            ack_requested=ack_requested, response_requested=response_requested,
            payload=payload)
        self.message_type = MSG_IDS[StateWifiInfo]

    def get_payload(self) -> bytes:
        self.payload_fields.append(("Signal (mW)", self.signal))
        self.payload_fields.append(("TX (bytes since on)", self.tx))
        self.payload_fields.append(("RX (bytes since on)", self.rx))
        self.payload_fields.append(("Reserved", self.reserved1))
        signal = little_endian(bitstring.pack("float:32", self.signal))
        tx = little_endian(bitstring.pack("uint:32", self.tx))
        rx = little_endian(bitstring.pack("uint:32", self.rx))
        reserved1 = little_endian(bitstring.pack("int:16", self.reserved1))
        payload = signal + tx + rx + reserved1
        return payload


class GetWifiFirmware(Message):
    def __init__(
            self, *, target_addr: str, source_id: int, seq_num: int,
            payload: Dict[str, Any]=None,
            ack_requested: bool=False, response_requested: bool=False) -> None:

        if payload is None:
            payload = {}

        super().__init__(
            target_addr=target_addr, source_id=source_id,
            seq_num=seq_num,
            ack_requested=ack_requested, response_requested=response_requested,
            payload=payload)
        self.message_type = MSG_IDS[GetWifiFirmware]


class StateWifiFirmware(Message):
    def __init__(
            self, *, target_addr: str, source_id: int, seq_num: int,
            payload: Dict[str, Any]=None,
            ack_requested: bool=False, response_requested: bool=False) -> None:

        if payload is None:
            payload = {}

        self.build = payload["build"]
        self.reserved1 = payload["reserved1"]
        self.version = payload["version"]
        super().__init__(
            target_addr=target_addr, source_id=source_id,
            seq_num=seq_num,
            ack_requested=ack_requested, response_requested=response_requested,
            payload=payload)
        self.message_type = MSG_IDS[StateWifiFirmware]

    def get_payload(self) -> bytes:
        self.payload_fields.append(("Timestamp of Build", self.build))
        self.payload_fields.append(("Reserved", self.reserved1))
        self.payload_fields.append(("Version", self.version))
        build = little_endian(bitstring.pack("uint:64", self.build))
        reserved1 = little_endian(bitstring.pack("uint:64", self.reserved1))
        version = little_endian(bitstring.pack("uint:32", self.version))
        payload = build + reserved1 + version
        return payload


class GetPower(Message):
    def __init__(
            self, *, target_addr: str, source_id: int, seq_num: int,
            payload: Dict[str, Any]=None,
            ack_requested: bool=False, response_requested: bool=False) -> None:

        if payload is None:
            payload = {}

        super().__init__(
            target_addr=target_addr, source_id=source_id,
            seq_num=seq_num,
            ack_requested=ack_requested, response_requested=response_requested,
            payload=payload)
        self.message_type = MSG_IDS[GetPower]


class SetPower(Message):
    def __init__(
            self, *, target_addr: str, source_id: int, seq_num: int,
            payload: Dict[str, Any]=None,
            ack_requested: bool=False, response_requested: bool=False) -> None:

        if payload is None:
            payload = {}

        self.power_level = payload["power_level"]
        super().__init__(
            target_addr=target_addr, source_id=source_id,
            seq_num=seq_num,
            ack_requested=ack_requested, response_requested=response_requested,
            payload=payload)
        self.message_type = MSG_IDS[SetPower]

    def get_payload(self) -> bytes:
        self.payload_fields.append(("Power", self.power_level))
        power_level = little_endian(bitstring.pack("uint:16", self.power_level))
        payload = power_level
        return payload


class StatePower(Message):
    def __init__(
            self, *, target_addr: str, source_id: int, seq_num: int,
            payload: Dict[str, Any]=None,
            ack_requested: bool=False, response_requested: bool=False) -> None:

        if payload is None:
            payload = {}

        self.power_level = payload["power_level"]
        super().__init__(
            target_addr=target_addr, source_id=source_id,
            seq_num=seq_num,
            ack_requested=ack_requested, response_requested=response_requested,
            payload=payload)
        self.message_type = MSG_IDS[StatePower]

    def get_payload(self) -> bytes:
        self.payload_fields.append(("Power", self.power_level))
        power_level = little_endian(bitstring.pack("uint:16", self.power_level))
        payload = power_level
        return payload


class GetLabel(Message):
    def __init__(
            self, *, target_addr: str, source_id: int, seq_num: int,
            payload: Dict[str, Any]=None,
            ack_requested: bool=False, response_requested: bool=False) -> None:

        if payload is None:
            payload = {}

        super().__init__(
            target_addr=target_addr, source_id=source_id,
            seq_num=seq_num,
            ack_requested=ack_requested, response_requested=response_requested,
            payload=payload)
        self.message_type = MSG_IDS[GetLabel]


class SetLabel(Message):
    def __init__(
            self, *, target_addr: str, source_id: int, seq_num: int,
            payload: Dict[str, Any]=None,
            ack_requested: bool=False, response_requested: bool=False) -> None:

        if payload is None:
            payload = {}

        self.label = payload["label"]
        super().__init__(
            target_addr=target_addr, source_id=source_id,
            seq_num=seq_num,
            ack_requested=ack_requested, response_requested=response_requested,
            payload=payload)
        self.message_type = MSG_IDS[SetLabel]

    def get_payload(self) -> bytes:
        self.payload_fields.append(("Label", self.label))
        field_len_bytes = 32
        label = b"".join(little_endian(bitstring.pack("uint:8", ord(c))) for c in self.label)
        padding = b"".join(little_endian(bitstring.pack("uint:8", 0)) for i in range(field_len_bytes - len(self.label)))
        payload = label + padding
        return payload


class StateLabel(Message):
    def __init__(
            self, *, target_addr: str, source_id: int, seq_num: int,
            payload: Dict[str, Any]=None,
            ack_requested: bool=False, response_requested: bool=False) -> None:

        if payload is None:
            payload = {}

        self.label = payload["label"]
        super().__init__(
            target_addr=target_addr, source_id=source_id,
            seq_num=seq_num,
            ack_requested=ack_requested, response_requested=response_requested,
            payload=payload)
        self.message_type = MSG_IDS[StateLabel]

    def get_payload(self) -> bytes:
        self.payload_fields.append(("Label", self.label))
        field_len_bytes = 32
        label = b"".join(little_endian(bitstring.pack("uint:8", c)) for c in self.label)
        padding = b"".join(little_endian(bitstring.pack("uint:8", 0)) for i in range(field_len_bytes - len(self.label)))
        payload = label + padding
        return payload


class GetVersion(Message):
    def __init__(
            self, *, target_addr: str, source_id: int, seq_num: int,
            payload: Dict[str, Any]=None,
            ack_requested: bool=False, response_requested: bool=False) -> None:

        if payload is None:
            payload = {}

        super().__init__(
            target_addr=target_addr, source_id=source_id,
            seq_num=seq_num,
            ack_requested=ack_requested, response_requested=response_requested,
            payload=payload)
        self.message_type = MSG_IDS[GetVersion]


class StateVersion(Message):
    def __init__(
            self, *, target_addr: str, source_id: int, seq_num: int,
            payload: Dict[str, Any]=None,
            ack_requested: bool=False, response_requested: bool=False) -> None:

        if payload is None:
            payload = {}

        self.vendor = payload["vendor"]
        self.product = payload["product"]
        self.version = payload["version"]
        super().__init__(
            target_addr=target_addr, source_id=source_id,
            seq_num=seq_num,
            ack_requested=ack_requested, response_requested=response_requested,
            payload=payload)
        self.message_type = MSG_IDS[StateVersion]

    def get_payload(self) -> bytes:
        self.payload_fields.append(("Vendor", self.vendor))
        self.payload_fields.append(("Reserved", self.product))
        self.payload_fields.append(("Version", self.version))
        vendor = little_endian(bitstring.pack("uint:32", self.vendor))
        product = little_endian(bitstring.pack("uint:32", self.product))
        version = little_endian(bitstring.pack("uint:32", self.version))
        payload = vendor + product + version
        return payload


class GetInfo(Message):
    def __init__(
            self, *, target_addr: str, source_id: int, seq_num: int,
            payload: Dict[str, Any]=None,
            ack_requested: bool=False, response_requested: bool=False) -> None:

        if payload is None:
            payload = {}

        super().__init__(
            target_addr=target_addr, source_id=source_id,
            seq_num=seq_num,
            ack_requested=ack_requested, response_requested=response_requested,
            payload=payload)
        self.message_type = MSG_IDS[GetInfo]


class StateInfo(Message):
    def __init__(
            self, *, target_addr: str, source_id: int, seq_num: int,
            payload: Dict[str, Any]=None,
            ack_requested: bool=False, response_requested: bool=False) -> None:

        if payload is None:
            payload = {}

        self.time = payload["time"]
        self.uptime = payload["uptime"]
        self.downtime = payload["downtime"]
        super().__init__(
            target_addr=target_addr, source_id=source_id,
            seq_num=seq_num,
            ack_requested=ack_requested, response_requested=response_requested,
            payload=payload)
        self.message_type = MSG_IDS[StateInfo]

    def get_payload(self) -> bytes:
        self.payload_fields.append(("Current Time", self.time))
        self.payload_fields.append(("Uptime (ns)", self.uptime))
        self.payload_fields.append(("Last Downtime Duration (ns) (5 second error)", self.downtime))
        time = little_endian(bitstring.pack("uint:64", self.time))
        uptime = little_endian(bitstring.pack("uint:64", self.uptime))
        downtime = little_endian(bitstring.pack("uint:64", self.downtime))
        payload = time + uptime + downtime
        return payload


class GetLocation(Message):
    def __init__(
            self, *, target_addr: str, source_id: int, seq_num: int,
            payload: Dict[str, Any]=None,
            ack_requested: bool=False, response_requested: bool=False) -> None:

        if payload is None:
            payload = {}

        super().__init__(
            target_addr=target_addr, source_id=source_id,
            seq_num=seq_num,
            ack_requested=ack_requested, response_requested=response_requested,
            payload=payload)
        self.message_type = MSG_IDS[GetLocation]


class StateLocation(Message):
    def __init__(
            self, *, target_addr: str, source_id: int, seq_num: int,
            payload: Dict[str, Any]=None,
            ack_requested: bool=False, response_requested: bool=False) -> None:

        if payload is None:
            payload = {}

        self.location = payload["location"]
        self.label = payload["label"]
        self.updated_at = payload["updated_at"]
        super().__init__(
            target_addr=target_addr, source_id=source_id,
            seq_num=seq_num,
            ack_requested=ack_requested, response_requested=response_requested,
            payload=payload)
        self.message_type = MSG_IDS[StateLocation]

    def get_payload(self) -> bytes:
        self.payload_fields.append(("Location ", self.location))
        self.payload_fields.append(("Label ", self.label))
        self.payload_fields.append(("Updated At ", self.updated_at))
        location = b"".join(little_endian(bitstring.pack("uint:8", b)) for b in self.location)
        label = b"".join(little_endian(bitstring.pack("uint:8", c)) for c in self.label)
        label_padding = b"".join(little_endian(bitstring.pack("uint:8", 0)) for i in range(32 - len(self.label)))
        label += label_padding
        updated_at = little_endian(bitstring.pack("uint:64", self.updated_at))
        payload = location + label + updated_at
        return payload


class GetGroup(Message):
    def __init__(
            self, *, target_addr: str, source_id: int, seq_num: int,
            payload: Dict[str, Any]=None,
            ack_requested: bool=False, response_requested: bool=False) -> None:

        if payload is None:
            payload = {}

        super().__init__(
            target_addr=target_addr, source_id=source_id,
            seq_num=seq_num,
            ack_requested=ack_requested, response_requested=response_requested,
            payload=payload)
        self.message_type = MSG_IDS[GetGroup]


class StateGroup(Message):
    def __init__(
            self, *, target_addr: str, source_id: int, seq_num: int,
            payload: Dict[str, Any]=None,
            ack_requested: bool=False, response_requested: bool=False) -> None:

        if payload is None:
            payload = {}

        self.group = payload["group"]
        self.label = payload["label"]
        self.updated_at = payload["updated_at"]
        super().__init__(
            target_addr=target_addr, source_id=source_id,
            seq_num=seq_num,
            ack_requested=ack_requested, response_requested=response_requested,
            payload=payload)
        self.message_type = MSG_IDS[StateGroup]

    def get_payload(self) -> bytes:
        self.payload_fields.append(("Group ", self.group))
        self.payload_fields.append(("Label ", self.label))
        self.payload_fields.append(("Updated At ", self.updated_at))
        group = b"".join(little_endian(bitstring.pack("uint:8", b)) for b in self.group)
        label = b"".join(little_endian(bitstring.pack("uint:8", c)) for c in self.label)
        label_padding = b"".join(little_endian(bitstring.pack("uint:8", 0)) for i in range(32 - len(self.label)))
        label += label_padding
        updated_at = little_endian(bitstring.pack("uint:64", self.updated_at))
        payload = group + label + updated_at
        return payload


class Acknowledgement(Message):
    def __init__(
            self, *, target_addr: str, source_id: int, seq_num: int,
            payload: Dict[str, Any]=None,
            ack_requested: bool=False, response_requested: bool=False) -> None:

        if payload is None:
            payload = {}

        super().__init__(
            target_addr=target_addr, source_id=source_id,
            seq_num=seq_num,
            ack_requested=ack_requested, response_requested=response_requested,
            payload=payload)
        self.message_type = MSG_IDS[Acknowledgement]


class EchoRequest(Message):
    def __init__(
            self, *, target_addr: str, source_id: int, seq_num: int,
            payload: Dict[str, Any]=None,
            ack_requested: bool=False, response_requested: bool=False) -> None:

        if payload is None:
            payload = {}

        self.byte_array = payload["byte_array"]
        super().__init__(
            target_addr=target_addr, source_id=source_id,
            seq_num=seq_num,
            ack_requested=ack_requested, response_requested=response_requested,
            payload=payload)
        self.message_type = MSG_IDS[EchoRequest]

    def get_payload(self) -> bytes:
        field_len = 64
        self.payload_fields.append(("Byte Array", self.byte_array))
        byte_array = b"".join(little_endian(bitstring.pack("uint:8", b)) for b in self.byte_array)
        byte_array_len = len(byte_array)
        if byte_array_len < field_len:
            byte_array += b"".join(little_endian(bitstring.pack("uint:8", 0))
                                   for i in range(field_len - byte_array_len))
        elif byte_array_len > field_len:
            byte_array = byte_array[:field_len]
        payload = byte_array
        return payload


class EchoResponse(Message):
    def __init__(
            self, *, target_addr: str, source_id: int, seq_num: int,
            payload: Dict[str, Any]=None,
            ack_requested: bool=False, response_requested: bool=False) -> None:

        if payload is None:
            payload = {}

        self.byte_array = payload["byte_array"]
        super().__init__(
            target_addr=target_addr, source_id=source_id,
            seq_num=seq_num,
            ack_requested=ack_requested, response_requested=response_requested,
            payload=payload)
        self.message_type = MSG_IDS[EchoResponse]

    def get_payload(self) -> bytes:
        self.payload_fields.append(("Byte Array", self.byte_array))
        byte_array = b"".join(little_endian(bitstring.pack("uint:8", b)) for b in self.byte_array)
        payload = byte_array
        return payload


# LIGHT MESSAGES


class LightGet(Message):
    def __init__(
            self, *, target_addr: str, source_id: int, seq_num: int,
            payload: Dict[str, Any]=None,
            ack_requested: bool=False, response_requested: bool=False) -> None:

        if payload is None:
            payload = {}

        super().__init__(
            target_addr=target_addr, source_id=source_id,
            seq_num=seq_num,
            ack_requested=ack_requested, response_requested=response_requested,
            payload=payload)
        self.message_type = MSG_IDS[LightGet]


class LightSetColor(Message):
    def __init__(
            self, *, target_addr: str, source_id: int, seq_num: int,
            payload: Dict[str, Any]=None,
            ack_requested: bool=False, response_requested: bool=False) -> None:

        if payload is None:
            payload = {}

        self.color = payload["color"]
        self.duration = payload["duration"]
        super().__init__(
            target_addr=target_addr, source_id=source_id,
            seq_num=seq_num,
            ack_requested=ack_requested, response_requested=response_requested,
            payload=payload)
        self.message_type = MSG_IDS[LightSetColor]

    def get_payload(self) -> bytes:
        reserved_8 = little_endian(bitstring.pack("uint:8", self.reserved))
        color = b"".join(little_endian(bitstring.pack("uint:16", field)) for field in self.color)
        duration = little_endian(bitstring.pack("uint:32", self.duration))
        payload = reserved_8 + color + duration
        # payloadUi = " ".join("{:02x}".format(c) for c in payload)
        return payload


class LightSetWaveform(Message):
    def __init__(
            self, *, target_addr: str, source_id: int, seq_num: int,
            payload: Dict[str, Any]=None,
            ack_requested: bool=False, response_requested: bool=False) -> None:

        if payload is None:
            payload = {}

        self.transient = payload["transient"]
        self.color = payload["color"]
        self.period = payload["period"]
        self.cycles = payload["cycles"]
        self.duty_cycle = payload["duty_cycle"]
        self.waveform = payload["waveform"]
        super().__init__(
            target_addr=target_addr, source_id=source_id,
            seq_num=seq_num,
            ack_requested=ack_requested, response_requested=response_requested,
            payload=payload)
        self.message_type = MSG_IDS[LightSetWaveform]

    def get_payload(self) -> bytes:
        reserved_8 = little_endian(bitstring.pack("uint:8", self.reserved))
        transient = little_endian(bitstring.pack("uint:8", self.transient))
        color = b"".join(little_endian(bitstring.pack("uint:16", field)) for field in self.color)
        period = little_endian(bitstring.pack("uint:32", self.period))
        cycles = little_endian(bitstring.pack("float:32", self.cycles))
        duty_cycle = little_endian(bitstring.pack("int:16", self.duty_cycle))
        waveform = little_endian(bitstring.pack("uint:8", self.waveform))
        payload = reserved_8 + transient + color + period + cycles + duty_cycle + waveform
        # payloadUi = " ".join("{:02x}".format(c) for c in payload)
        return payload


class LightState(Message):
    def __init__(
            self, *, target_addr: str, source_id: int, seq_num: int,
            payload: Dict[str, Any]=None,
            ack_requested: bool=False, response_requested: bool=False) -> None:

        if payload is None:
            payload = {}

        self.color = payload["color"]
        self.reserved1 = payload["reserved1"]
        self.power_level = payload["power_level"]
        self.label = payload["label"]
        self.reserved2 = payload["reserved2"]
        super().__init__(
            target_addr=target_addr, source_id=source_id,
            seq_num=seq_num,
            ack_requested=ack_requested, response_requested=response_requested,
            payload=payload)
        self.message_type = MSG_IDS[LightState]

    def get_payload(self) -> bytes:
        self.payload_fields.append(("Color (HSBK)", self.color))
        self.payload_fields.append(("Reserved", self.reserved1))
        self.payload_fields.append(("Power Level", self.power_level))
        self.payload_fields.append(("Label", self.label))
        self.payload_fields.append(("Reserved", self.reserved2))
        color = b"".join(little_endian(bitstring.pack("uint:16", field)) for field in self.color)
        reserved1 = little_endian(bitstring.pack("int:16", self.reserved1))
        power_level = little_endian(bitstring.pack("uint:16", self.power_level))
        label = b"".join(little_endian(bitstring.pack("uint:8", c)) for c in self.label)
        label_padding = b"".join(little_endian(bitstring.pack("uint:8", 0)) for i in range(32 - len(self.label)))
        label += label_padding
        reserved2 = little_endian(bitstring.pack("uint:64", self.reserved1))
        payload = color + reserved1 + power_level + label + reserved2
        return payload


class LightGetPower(Message):
    def __init__(
            self, *, target_addr: str, source_id: int, seq_num: int,
            payload: Dict[str, Any]=None,
            ack_requested: bool=False, response_requested: bool=False) -> None:

        if payload is None:
            payload = {}

        super().__init__(
            target_addr=target_addr, source_id=source_id,
            seq_num=seq_num,
            ack_requested=ack_requested, response_requested=response_requested,
            payload=payload)
        self.message_type = MSG_IDS[LightGetPower]


class LightSetPower(Message):
    def __init__(
            self, *, target_addr: str, source_id: int, seq_num: int,
            payload: Dict[str, Any]=None,
            ack_requested: bool=False, response_requested: bool=False) -> None:

        if payload is None:
            payload = {}

        self.power_level = payload["power_level"]
        self.duration = payload["duration"]
        super().__init__(
            target_addr=target_addr, source_id=source_id,
            seq_num=seq_num,
            ack_requested=ack_requested, response_requested=response_requested,
            payload=payload)
        self.message_type = MSG_IDS[LightSetPower]

    def get_payload(self) -> bytes:
        power_level = little_endian(bitstring.pack("uint:16", self.power_level))
        duration = little_endian(bitstring.pack("uint:32", self.duration))
        payload = power_level + duration
        return payload


class LightStatePower(Message):
    def __init__(
            self, *, target_addr: str, source_id: int, seq_num: int,
            payload: Dict[str, Any]=None,
            ack_requested: bool=False, response_requested: bool=False) -> None:

        if payload is None:
            payload = {}

        self.power_level = payload["power_level"]
        super().__init__(
            target_addr=target_addr, source_id=source_id,
            seq_num=seq_num,
            ack_requested=ack_requested, response_requested=response_requested,
            payload=payload)
        self.message_type = MSG_IDS[LightStatePower]

    def get_payload(self) -> bytes:
        self.payload_fields.append(("Power Level", self.power_level))
        power_level = little_endian(bitstring.pack("uint:16", self.power_level))
        payload = power_level
        return payload

# INFRARED MESSAGES


class LightGetInfrared(Message):
    def __init__(
            self, *, target_addr: str, source_id: int, seq_num: int,
            payload: Dict[str, Any]=None,
            ack_requested: bool=False, response_requested: bool=False) -> None:

        if payload is None:
            payload = {}

        super().__init__(
            target_addr=target_addr, source_id=source_id,
            seq_num=seq_num,
            ack_requested=ack_requested, response_requested=response_requested,
            payload=payload)
        self.message_type = MSG_IDS[LightGetInfrared]


class LightStateInfrared(Message):
    def __init__(
            self, *, target_addr: str, source_id: int, seq_num: int,
            payload: Dict[str, Any]=None,
            ack_requested: bool=False, response_requested: bool=False) -> None:

        if payload is None:
            payload = {}

        self.infrared_brightness = payload["infrared_brightness"]
        super().__init__(
            target_addr=target_addr, source_id=source_id,
            seq_num=seq_num,
            ack_requested=ack_requested, response_requested=response_requested,
            payload=payload)
        self.message_type = MSG_IDS[LightStateInfrared]

    def get_payload(self) -> bytes:
        self.payload_fields.append(("Infrared Brightness", self.infrared_brightness))
        infrared_brightness = little_endian(bitstring.pack("uint:16", self.infrared_brightness))
        payload = infrared_brightness
        return payload


class LightSetInfrared(Message):
    def __init__(
            self, *, target_addr: str, source_id: int, seq_num: int,
            payload: Dict[str, Any]=None,
            ack_requested: bool=False, response_requested: bool=False) -> None:

        if payload is None:
            payload = {}

        self.infrared_brightness = payload["infrared_brightness"]
        super().__init__(
            target_addr=target_addr, source_id=source_id,
            seq_num=seq_num,
            ack_requested=ack_requested, response_requested=response_requested,
            payload=payload)
        self.message_type = MSG_IDS[LightSetInfrared]

    def get_payload(self) -> bytes:
        infrared_brightness = little_endian(bitstring.pack("uint:16", self.infrared_brightness))
        payload = infrared_brightness
        return payload

# MULTIZONE MESSAGES


class MultiZoneStateMultiZone(Message):
    def __init__(
            self, *, target_addr: str, source_id: int, seq_num: int,
            payload: Dict[str, Any]=None,
            ack_requested: bool=False, response_requested: bool=False) -> None:

        if payload is None:
            payload = {}

        self.count = payload["count"]
        self.index = payload["index"]
        self.color = payload["color"]
        super().__init__(
            target_addr=target_addr, source_id=source_id,
            seq_num=seq_num,
            ack_requested=ack_requested, response_requested=response_requested,
            payload=payload)
        self.message_type = MSG_IDS[MultiZoneStateMultiZone]

    def get_payload(self) -> bytes:
        self.payload_fields.append(("Count", self.count))
        self.payload_fields.append(("Index", self.index))
        self.payload_fields.append(("Color (HSBK)", self.color))
        count = little_endian(bitstring.pack("uint:8", self.count))
        index = little_endian(bitstring.pack("uint:8", self.index))
        payload = count + index
        for color in self.color:
            payload += b"".join(little_endian(bitstring.pack("uint:16", field)) for field in color)
        return payload


class MultiZoneStateZone(Message):  # 503
    def __init__(
            self, *, target_addr: str, source_id: int, seq_num: int,
            payload: Dict[str, Any]=None,
            ack_requested: bool=False, response_requested: bool=False) -> None:

        if payload is None:
            payload = {}

        self.count = payload["count"]
        self.index = payload["index"]
        self.color = payload["color"]
        super().__init__(
            target_addr=target_addr, source_id=source_id,
            seq_num=seq_num,
            ack_requested=ack_requested, response_requested=response_requested,
            payload=payload)
        self.message_type = MSG_IDS[MultiZoneStateZone]

    def get_payload(self) -> bytes:
        self.payload_fields.append(("Count", self.count))
        self.payload_fields.append(("Index", self.index))
        self.payload_fields.append(("Color (HSBK)", self.color))
        count = little_endian(bitstring.pack("uint:8", self.count))
        index = little_endian(bitstring.pack("uint:8", self.index))
        color = b"".join(little_endian(bitstring.pack("uint:16", field)) for field in self.color)
        payload = count + index + color
        return payload


class MultiZoneSetColorZones(Message):
    def __init__(
            self, *, target_addr: str, source_id: int, seq_num: int,
            payload: Dict[str, Any]=None,
            ack_requested: bool=False, response_requested: bool=False) -> None:

        if payload is None:
            payload = {}

        self.start_index = payload["start_index"]
        self.end_index = payload["end_index"]
        self.color = payload["color"]
        self.duration = payload["duration"]
        self.apply = payload["apply"]
        super().__init__(
            target_addr=target_addr, source_id=source_id,
            seq_num=seq_num,
            ack_requested=ack_requested, response_requested=response_requested,
            payload=payload)
        self.message_type = MSG_IDS[MultiZoneSetColorZones]

    def get_payload(self) -> bytes:
        start_index = little_endian(bitstring.pack("uint:8", self.start_index))
        end_index = little_endian(bitstring.pack("uint:8", self.end_index))
        color = b"".join(little_endian(bitstring.pack("uint:16", field)) for field in self.color)
        duration = little_endian(bitstring.pack("uint:32", self.duration))
        apply = little_endian(bitstring.pack("uint:8", self.apply))
        payload = start_index + end_index + color + duration + apply
        return payload


class MultiZoneGetColorZones(Message):
    def __init__(
            self, *, target_addr: str, source_id: int, seq_num: int,
            payload: Dict[str, Any]=None,
            ack_requested: bool=False, response_requested: bool=False) -> None:

        if payload is None:
            payload = {}

        self.start_index = payload["start_index"]
        self.end_index = payload["end_index"]
        super().__init__(
            target_addr=target_addr, source_id=source_id,
            seq_num=seq_num,
            ack_requested=ack_requested, response_requested=response_requested,
            payload=payload)
        self.message_type = MSG_IDS[MultiZoneGetColorZones]

    def get_payload(self) -> bytes:
        start_index = little_endian(bitstring.pack("uint:8", self.start_index))
        end_index = little_endian(bitstring.pack("uint:8", self.end_index))
        payload = start_index + end_index
        return payload


MSG_IDS = {GetService: 2,
           StateService: 3,
           GetHostInfo: 12,
           StateHostInfo: 13,
           GetHostFirmware: 14,
           StateHostFirmware: 15,
           GetWifiInfo: 16,
           StateWifiInfo: 17,
           GetWifiFirmware: 18,
           StateWifiFirmware: 19,
           GetPower: 20,
           SetPower: 21,
           StatePower: 22,
           GetLabel: 23,
           SetLabel: 24,
           StateLabel: 25,
           GetVersion: 32,
           StateVersion: 33,
           GetInfo: 34,
           StateInfo: 35,
           Acknowledgement: 45,
           GetLocation: 48,
           StateLocation: 50,
           GetGroup: 51,
           StateGroup: 53,
           EchoRequest: 58,
           EchoResponse: 59,
           LightGet: 101,
           LightSetColor: 102,
           LightSetWaveform: 103,
           LightState: 107,
           LightGetPower: 116,
           LightSetPower: 117,
           LightStatePower: 118,
           LightGetInfrared: 120,
           LightStateInfrared: 121,
           LightSetInfrared: 122,
           MultiZoneSetColorZones: 501,
           MultiZoneGetColorZones: 502,
           MultiZoneStateZone: 503,
           MultiZoneStateMultiZone: 506}

SERVICE_IDS = {1: "UDP",
               2: "reserved",
               3: "reserved",
               4: "reserved"}

STR_MAP = {65535: "On",
           0: "Off",
           None: "Unknown"}

ZONE_MAP = {0: "NO_APPLY",
            1: "APPLY",
            2: "APPLY_ONLY"}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    "bitstring",
]

setup_requirements = [
    'pytest-runner',
]

test_requirements = [
    'pytest',
]

setup(
    name='aiolifxc',
    version='0.5.3',
    description='API for local communication with LIFX devices '
                'over a LAN with asyncio.',
    long_description=readme + '\n\n' + history,
    author="Brian May",
    author_email='brian@linuxpenguins.xyz',
    url='http://github.com/brianmay/aiolifx',
    packages=find_packages(include=['aiolifxc']),
    include_package_data=True,
    install_requires=requirements,
    license='MIT',
    keywords=['lifx', 'light', 'automation'],
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    test_suite='tests',
    tests_require=test_requirements,
    setup_requires=setup_requirements,
)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# This application is an example on how to use aiolifx
#
# Copyright (c) 2016 François Wautier
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies
# or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
# IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
import sys
import asyncio as aio
from typing import List, Optional  # noqa

import aiolifxc as alix

UDP_BROADCAST_PORT = 56700

""" Simple light control from console. """


class MyDevices(alix.Devices):
    """ A simple class with a register and  unregister methods. """

    def __init__(self, loop: aio.AbstractEventLoop) -> None:
        super().__init__(loop=loop)
        self.selected = None  # type: Optional[alix.Devices]


def read_in(*, loop: aio.AbstractEventLoop, devices: MyDevices) -> None:
    """Reading from stdin and displaying menu"""

    selection = sys.stdin.readline().strip("\n")
    loop.create_task(read_in_process(selection=selection, devices=devices))


async def read_in_process(*, selection: str, devices: MyDevices) -> None:
    selected_lights = None   # type: Optional[alix.Lights]
    if devices.selected is not None:
        selected_lights = devices.selected.get_clone(alix.Lights)

    device_list = list(devices.device_list)  # type: List[alix.Device]
    device_list.sort(key=lambda x: x.mac_addr)

    lov = [x for x in selection.split(" ") if x != ""]
    if lov:
        if devices.selected is not None and selected_lights is not None:
            # try:
            if True:
                if int(lov[0]) == 0:
                    devices.selected = None
                elif int(lov[0]) == 1:
                    if len(lov) > 1:
                        try:
                            await devices.selected.set_power(lov[1].lower() in ["1", "on", "true"])
                            devices.selected = None
                        except alix.DeviceOffline:
                            print("Error: Device is offline")
                    else:
                        print("Error: For power you must indicate on or off\n")
                elif int(lov[0]) == 2:
                    if len(lov) > 2:
                        try:
                            color = alix.Color(
                                hue=0,
                                saturation=0,
                                brightness=int(lov[1]),
                                kelvin=int(lov[2]),
                            )
                            await selected_lights.set_color(color)

                            devices.selected = None
                        except (IndexError, ValueError):
                            print("Error: For white brightness (0-100) and temperature (2500-9000) must be numbers.\n")
                    else:
                        print("Error: For white you must indicate brightness (0-100) and temperature (2500-9000)\n")
                elif int(lov[0]) == 3:
                    if len(lov) > 3:
                        try:
                            color = alix.Color(
                                hue=int(lov[1]),
                                saturation=int(lov[2]),
                                brightness=int(lov[3]),
                                kelvin=3500,
                            )
                            await selected_lights.set_color(color)
                            devices.selected = None
                        except (IndexError, ValueError):
                            print("Error: For colour hue (0-360), "
                                  "saturation (0-100) and brightness (0-100) "
                                  "must be numbers.\n")
                    else:
                        print("Error: For colour you must indicate hue (0-360), "
                              "saturation (0-100) and brightness (0-100)\n")

                elif int(lov[0]) == 4:
                    for device in device_list:
                        await device.get_power()
                        await device.get_group()
                        await device.get_location()
                        await device.get_version()
                        print(device.device_characteristics_str("    "))
                        print(device.device_product_str("    "))
                    devices.selected = None
                elif int(lov[0]) == 5:
                    for device in device_list:
                        await device.get_host_firmware()
                        await device.get_wifi_firmware()
                        print(device.device_firmware_str("   "))
                    devices.selected = None
                elif int(lov[0]) == 6:
                    for device in device_list:
                        wifi_info = await device.get_wifi_info()
                        print(device.device_radio_str(wifi_info))
                    devices.selected = None
                elif int(lov[0]) == 7:
                    for device in device_list:
                        host_info = await device.get_host_info()
                        print(device.device_time_str(host_info))
                    devices.selected = None
                elif int(lov[0]) == 8:
                    if len(lov) > 3:
                        try:
                            color = alix.Color(
                                hue=int(lov[1]),
                                saturation=int(lov[2]),
                                brightness=int(lov[3]),
                                kelvin=3500,
                            )
                            await selected_lights.set_waveform(
                                color=color,
                                transient=1,
                                period=100,
                                cycles=30,
                                duty_cycle=0,
                                waveform=0
                            )
                            devices.selected = None
                        except (IndexError, ValueError):
                            print("Error: For pulse hue (0-360), "
                                  "saturation (0-100) and brightness (0-100) "
                                  "must be numbers.\n")
                    else:
                        print("Error: For pulse you must indicate hue (0-360),"
                              "saturation (0-100) and brightness (0-100))\n")
        else:

            if lov[0] == 'group':
                devices.selected = devices.get_by_group(lov[1])
            elif lov[0] == 'label':
                devices.selected = devices.get_by_label(lov[1])
            else:
                try:

                    index = int(lov[0])
                    if index > 0:
                        if index <= len(device_list):
                            devices.selected = devices.get_by_mac_addr(device_list[index-1].mac_addr)
                        else:
                            print("\nError: Not a valid selection.\n")

                except (IndexError, ValueError):
                    print("\nError: Selection must be a number.\n")

    if devices.selected:
        print("Select Function for {}:".format(", ".join([str(d) for d in device_list])))
        print("\t[1]\tPower (0 or 1)")
        print("\t[2]\tWhite (Brightness Temperature)")
        print("\t[3]\tColour (Hue Saturation Value)")
        print("\t[4]\tInfo")
        print("\t[5]\tFirmware")
        print("\t[6]\tWifi")
        print("\t[7]\tUptime")
        print("\t[8]\tPulse")
        print("")
        print("\t[0]\tBack to light selection")
    else:
        idx = 1
        print("Select Bulb:")
        for x in device_list:
            print("\t[{}]\t{}\t{}\t{}".format(idx, x.label, x.mac_addr, x.group))
            idx += 1
        print("")
        print("Alternatively type 'group <group>' or 'label <label>' to select a number of lights.")

    print("")
    print("Your choice: ", end='', flush=True)


def main() -> None:
    """ Main CLI function. """
    loop = aio.get_event_loop()
    devices = MyDevices(loop=loop)

    def read_in_wrapper() -> None:
        read_in(loop=loop, devices=devices)

    server = devices.start_discover()  # type: aio.Task
    loop.add_reader(sys.stdin, read_in_wrapper)

    try:
        print("Hit \"Enter\" to start")
        print("Use Ctrl-C to quit")
        loop.run_forever()
    except Exception as e:
        print("Got exception %s" % e)
    finally:
        server.cancel()
        loop.remove_reader(sys.stdin)
        loop.close()


main()
